# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.
import logging
import traceback
from typing import Dict, List

import pandas as pd

from pandapower.toolbox.grid_modification import fuse_buses
from pandapower.run import runpp
from pandapower.create import create_empty_network
from pandapower.auxiliary import pandapowerNet
from .convert_measurements import CreateMeasurements
from .. import cim_classes
from .. import cim_tools
from .. import pp_tools
from ..other_classes import ReportContainer, Report, LogLevel, ReportCode
from pandapower.control.util.auxiliary import create_q_capability_curve_characteristics_object

logger = logging.getLogger('cim.cim2pp.build_pp_net')

sc = cim_tools.get_pp_net_special_columns_dict()


class CimConverter:

    def __init__(self, cim_parser: cim_classes.CimParser, converter_classes: Dict, **kwargs):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cim_parser: cim_classes.CimParser = cim_parser
        self.kwargs = kwargs
        self.cim: Dict[str, Dict[str, pd.DataFrame]] = self.cim_parser.get_cim_dict()
        self.net: pandapowerNet = create_empty_network()
        self.bus_merge: pd.DataFrame = pd.DataFrame()
        self.power_trafo2w: pd.DataFrame = pd.DataFrame()
        self.power_trafo3w: pd.DataFrame = pd.DataFrame()
        self.report_container: ReportContainer = cim_parser.get_report_container()
        self.classes_dict = converter_classes

    def merge_eq_ssh_profile(self, cim_type: str, add_cim_type_column: bool = False) -> pd.DataFrame:
        return self.merge_eq_other_profiles(['ssh'], cim_type, add_cim_type_column)

    def merge_eq_sc_profile(self, cim_type: str, add_cim_type_column: bool = False) -> pd.DataFrame:
        return self.merge_eq_other_profiles(['sc'], cim_type, add_cim_type_column)

    def merge_eq_other_profiles(self, other_profiles: List[str], cim_type: str,
                                add_cim_type_column: bool = False) -> pd.DataFrame:
        df = self.cim['eq'][cim_type]
        for other_profile in other_profiles:
            if cim_type not in self.cim[other_profile].keys():
                self.logger.debug("No entries found in %s profile for cim object %s", other_profile, cim_type)
                return self.cim['eq'][cim_type].copy()
            df = pd.merge(df, self.cim[other_profile][cim_type], how='left', on='rdfId')
        if add_cim_type_column:
            df[sc['o_cl']] = cim_type
        return df

    def copy_to_pp(self, pp_type: str, input_df: pd.DataFrame):
        self.logger.debug("Copy %s datasets to pandapower network with type %s" % (input_df.index.size, pp_type))
        if pp_type not in self.net.keys():
            self.logger.warning("Missing pandapower type %s in the pandapower network!" % pp_type)
            self.report_container.add_log(Report(
                level=LogLevel.WARNING, code=ReportCode.WARNING_CONVERTING,
                message="Missing pandapower type %s in the pandapower network!" % pp_type))
            return
        self.net[pp_type] = pd.concat([self.net[pp_type],
                                      input_df[list(set(self.net[pp_type].columns).intersection(input_df.columns))]],
                                      ignore_index=True, sort=False)

    # noinspection PyShadowingNames
    def convert_to_pp(
            self, convert_line_to_switch: bool = False, line_r_limit: float = 0.1, line_x_limit: float = 0.1, **kwargs
    ) -> pandapowerNet:
        """
        Build the pandapower net.

        :param convert_line_to_switch: Set this parameter to True to enable line -> switch conversion. All lines with a
        resistance lower or equal than line_r_limit or a reactance lower or equal than line_x_limit will become a
        switch. Optional, default: False
        :param line_r_limit: The limit from resistance. Optional, default: 0.1
        :param line_x_limit: The limit from reactance. Optional, default: 0.1
        :return: The pandapower net.
        """
        self.logger.info("Start building the pandapower net.")
        self.report_container.add_log(Report(level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
                                             message="Start building the pandapower net."))

        # create the empty pandapower net and add the additional columns
        self.net = cim_tools.extend_pp_net_cim(self.net, override=False)

        if 'sn_mva' in kwargs.keys():
            self.net['sn_mva'] = kwargs.get('sn_mva')

        # add the CIM IDs to the pandapower network
        for one_prf, one_profile_dict in self.cim.items():
            if 'FullModel' in one_profile_dict.keys() and one_profile_dict['FullModel'].index.size > 0:
                self.net['CGMES'][one_prf] = one_profile_dict['FullModel'].set_index('rdfId').to_dict(orient='index')
        # store the BaseVoltage IDs
        self.net['CGMES']['BaseVoltage'] = \
            pd.concat([self.cim['eq']['BaseVoltage'], self.cim['eq_bd']['BaseVoltage']],
                      sort=False, ignore_index=True)[['rdfId', 'nominalVoltage']]

        # --------- convert busses ---------
        self.classes_dict['ConnectivityNodesCim16'](cimConverter=self).convert_connectivity_nodes_cim16()
        # --------- convert external networks ---------
        self.classes_dict['externalNetworkInjectionsCim16'](
            cimConverter=self).convert_external_network_injections_cim16()
        # --------- convert lines ---------
        self.classes_dict['acLineSegmentsCim16'](cimConverter=self).convert_ac_line_segments_cim16(
            convert_line_to_switch, line_r_limit, line_x_limit)
        self.classes_dict['dcLineSegmentsCim16'](cimConverter=self).convert_dc_line_segments_cim16()
        # --------- convert switches ---------
        self.classes_dict['switchesCim16'](cimConverter=self).convert_switches_cim16()
        # --------- convert loads ---------
        self.classes_dict['energyConcumersCim16'](cimConverter=self).convert_energy_consumers_cim16()  #todo  correction of spelling mistake
        self.classes_dict['conformLoadsCim16'](cimConverter=self).convert_conform_loads_cim16()
        self.classes_dict['nonConformLoadsCim16'](cimConverter=self).convert_non_conform_loads_cim16()
        self.classes_dict['stationSuppliesCim16'](cimConverter=self).convert_station_supplies_cim16()
        # --------- convert generators ---------
        self.classes_dict['synchronousMachinesCim16'](cimConverter=self).convert_synchronous_machines_cim16()
        self.classes_dict['asynchronousMachinesCim16'](cimConverter=self).convert_asynchronous_machines_cim16()
        self.classes_dict['energySourcesCim16'](cimConverter=self).convert_energy_sources_cim16()
        # --------- convert shunt elements ---------
        self.classes_dict['linearShuntCompensatorCim16'](cimConverter=self).convert_linear_shunt_compensator_cim16()
        self.classes_dict['nonLinearShuntCompensatorCim16'](
            cimConverter=self).convert_nonlinear_shunt_compensator_cim16()
        self.classes_dict['staticVarCompensatorCim16'](cimConverter=self).convert_static_var_compensator_cim16()
        # --------- convert impedance elements ---------
        self.classes_dict['equivalentBranchesCim16'](cimConverter=self).convert_equivalent_branches_cim16()
        self.classes_dict['seriesCompensatorsCim16'](cimConverter=self).convert_series_compensators_cim16()
        # --------- convert extended ward and ward elements ---------
        self.classes_dict['equivalentInjectionsCim16'](cimConverter=self).convert_equivalent_injections_cim16()
        # --------- convert transformers ---------
        self.classes_dict['powerTransformersCim16'](cimConverter=self).convert_power_transformers_cim16()

        # --------- create reactive power capability characteristics ---------
        create_q_capability_curve_characteristics_object(self.net)

        # create the geo coordinates
        if self.cim['gl']['Location'].index.size > 0 and self.cim['gl']['PositionPoint'].index.size > 0:
            try:
                self.classes_dict['geoCoordinatesFromGLCim16'](cimConverter=self).add_geo_coordinates_from_gl_cim16()
            except Exception as e:
                self.logger.warning("Creating the geo coordinates failed, returning the net without geo coordinates!")
                self.logger.exception(e)
                self.report_container.add_log(Report(
                    level=LogLevel.WARNING, code=ReportCode.WARNING_CONVERTING,
                    message="Creating the geo coordinates failed, returning the net without geo coordinates!"))
                self.report_container.add_log(Report(
                    level=LogLevel.EXCEPTION, code=ReportCode.EXCEPTION_CONVERTING,
                    message=traceback.format_exc()))
        if self.cim['dl']['Diagram'].index.size > 0 and self.cim['dl']['DiagramObject'].index.size > 0 and \
                self.cim['dl']['DiagramObjectPoint'].index.size > 0:
            try:
                self.classes_dict['coordinatesFromDLCim16'](cimConverter=self).add_coordinates_from_dl_cim16(
                    diagram_name=kwargs.get('diagram_name', None))
            except Exception as e:
                self.logger.warning("Creating the coordinates failed, returning the net without coordinates!")
                self.logger.exception(e)
                self.report_container.add_log(Report(
                    level=LogLevel.WARNING, code=ReportCode.WARNING_CONVERTING,
                    message="Creating the coordinates failed, returning the net without coordinates!"))
                self.report_container.add_log(Report(level=LogLevel.EXCEPTION, code=ReportCode.EXCEPTION_CONVERTING,
                                                     message=traceback.format_exc()))
        self.net = pp_tools.set_pp_col_types(net=self.net)

        # create transformer tap controller
        self.classes_dict['tapController'](cimConverter=self).create_tap_controller_for_power_transformers()

        self.logger.info("Running a power flow.")
        self.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO, message="Running a power flow."))
        if kwargs.get('run_powerflow', False):
            try:
                runpp(self.net)
            except Exception as e:
                self.logger.error("Failed running a powerflow.")
                self.logger.exception(e)
                self.report_container.add_log(Report(
                    level=LogLevel.ERROR, code=ReportCode.ERROR, message="Failed running a powerflow."))
                self.report_container.add_log(Report(level=LogLevel.EXCEPTION, code=ReportCode.EXCEPTION,
                                                     message=traceback.format_exc()))
                if not kwargs.get('ignore_errors', True):
                    raise e
            else:
                self.logger.info("Power flow solved normal.")
                self.report_container.add_log(Report(
                    level=LogLevel.INFO, code=ReportCode.INFO, message="Power flow solved normal."))
        try:
            create_measurements = kwargs.get('create_measurements', None)
            if create_measurements is not None and create_measurements.lower() == 'sv':
                CreateMeasurements(self.net, self.cim).create_measurements_from_sv()
            elif create_measurements is not None and create_measurements.lower() == 'analog':
                CreateMeasurements(self.net, self.cim).create_measurements_from_analog()
            elif create_measurements is not None:
                self.report_container.add_log(Report(
                    level=LogLevel.ERROR, code=ReportCode.ERROR_CONVERTING,
                    message="Not supported value for argument 'create_measurements', check method signature for"
                            "valid values!"))
                raise ValueError("Not supported value for argument 'create_measurements', check method signature for"
                                 "valid values!")
        except Exception as e:
            self.logger.error("Creating the measurements failed, returning the net without measurements!")
            self.logger.exception(e)
            self.report_container.add_log(Report(
                level=LogLevel.ERROR, code=ReportCode.ERROR_CONVERTING,
                message="Creating the measurements failed, returning the net without measurements!"))
            self.report_container.add_log(Report(
                level=LogLevel.EXCEPTION, code=ReportCode.EXCEPTION_CONVERTING,
                message=traceback.format_exc()))
            self.net.measurement = self.net.measurement[0:0]
            if not kwargs.get('ignore_errors', True):
                raise e
        # a special fix for BB and NB mixed networks:
        # fuse boundary ConnectivityNodes with their TopologicalNodes
        bus_t = self.net.bus.reset_index(level=0, drop=False)
        bus_drop = bus_t.loc[bus_t[sc['o_prf']] == 'eq_bd', ['index', sc['o_id'], 'cim_topnode']]
        bus_drop = bus_drop.rename(columns={'index': 'b1'})
        bus_drop = pd.merge(bus_drop, bus_t[['index', sc['o_id']]].rename(columns={'index': 'b2', sc['o_id']: 'o_id2'}),
                            how='inner', left_on='cim_topnode', right_on='o_id2')
        if bus_drop.index.size > 0:
            for b1, b2 in bus_drop[['b1', 'b2']].itertuples(index=False):
                self.logger.info("Fusing buses: b1: %s, b2: %s" % (b1, b2))
                fuse_buses(self.net, b1, b2, drop=True, fuse_bus_measurements=True)
        # finally a fix for EquivalentInjections: If an EquivalentInjection is attached to boundary node, check if the
        # network behind this boundary node is attached. In this case, disable the EquivalentInjection.
        ward_t = self.net.ward.copy()
        ward_t['bus_prf'] = ward_t['bus'].map(self.net.bus[[sc['o_prf']]].to_dict().get(sc['o_prf']))
        self.net.ward.loc[(self.net.ward.bus.duplicated(keep=False) &
                           ((ward_t['bus_prf'] == 'eq_bd') | (ward_t['bus_prf'] == 'tp_bd'))), 'in_service'] = False
        xward_t = self.net.xward.copy()
        xward_t['bus_prf'] = xward_t['bus'].map(self.net.bus[[sc['o_prf']]].to_dict().get(sc['o_prf']))
        self.net.xward.loc[(self.net.xward.bus.duplicated(keep=False) &
                            ((xward_t['bus_prf'] == 'eq_bd') | (xward_t['bus_prf'] == 'tp_bd'))), 'in_service'] = False
        self.net['report_container'] = self.report_container
        return self.net



import logging
import time

import pandas as pd

from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net
from pandapower.converter.cim.other_classes import Report, LogLevel, ReportCode

logger = logging.getLogger('cim.cim2pp.converter_classes.stationSuppliesCim16')

sc = cim_tools.get_pp_net_special_columns_dict()


class StationSuppliesCim16:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def convert_station_supplies_cim16(self):
        time_start = time.time()
        self.logger.info("Start converting StationSupplies.")
        eqssh_station_supplies = self._prepare_station_supplies_cim16()
        self.cimConverter.copy_to_pp('load', eqssh_station_supplies)
        self.logger.info("Created %s loads in %ss." % (eqssh_station_supplies.index.size, time.time() - time_start))
        self.cimConverter.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
            message="Created %s loads from StationSupplies in %ss." %
                    (eqssh_station_supplies.index.size, time.time() - time_start)))

    def _prepare_station_supplies_cim16(self) -> pd.DataFrame:
        eqssh_station_supplies = self.cimConverter.merge_eq_ssh_profile('StationSupply', add_cim_type_column=True)
        eqssh_station_supplies = pd.merge(eqssh_station_supplies, self.cimConverter.bus_merge, how='left', on='rdfId')
        if 'inService' in eqssh_station_supplies.columns:
            eqssh_station_supplies['connected'] = eqssh_station_supplies['connected'] & eqssh_station_supplies['inService']
        eqssh_station_supplies = eqssh_station_supplies.rename(columns={'rdfId': sc['o_id'], 'rdfId_Terminal': sc['t'], 'index_bus': 'bus',
                                               'connected': 'in_service', 'p': 'p_mw', 'q': 'q_mvar'})
        eqssh_station_supplies['const_i_percent'] = 0.
        eqssh_station_supplies['const_z_percent'] = 0.
        eqssh_station_supplies['scaling'] = 1.
        eqssh_station_supplies['type'] = None
        return eqssh_station_supplies



# -*- coding: utf-8 -*-

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

"""
Defines constants for named column indices to DC bus matrix.

Some examples of usage, after defining the constants using the line above,
are::

    Pd = bus_dc[3, DC_PD]     # get the real power demand at DC bus 4
    bus_dc[:, DC_VMIN] = 0.95 # set the min voltage magnitude to 0.95 at all DC buses

The index, name, and meaning of each column of the DC bus matrix is given below:

columns 0-10 must be included in the input matrix (bus_dc in ppc dict)
    0.  C{DC_BUS_I}       DC bus number (1 to 29997)
    1.  C{DC_BUS_TYPE}    DC bus type (0 = reference, 1 = P, 2 = isolated)
    2.  C{DC_PD}          real power demand (MW)
    3.  C{DC_GS}          shunt conductance (MW at V = 1.0 p.u.)
    4.  C{DC_BUS_AREA}    area number, 1-100
    5.  C{DC_VM}          voltage magnitude (p.u.)
    6.  C{DC_BASE_KV}     base voltage (kV)
    7.  C{DC_ZONE}        loss zone (1-999)
    8.  C{DC_VMAX}        maximum voltage magnitude (p.u.)
    9.  C{DC_VMIN}        minimum voltage magnitude (p.u.)
    10. C{DC_SL_FAC}      scaling factor for special purposes

additional constants, used to assign/compare values in the C{DC_BUS_TYPE} column
    0.  C{DC_REF}   reference DC bus
    1.  C{DC_P}     DC bus with active power demand
    2.  C{DC_NONE}  DC bus that is out of service

"""

# define DC bus types
DC_REF = 0
DC_B2B = 1
DC_P = 2
DC_NONE = 3

# define the indices
DC_BUS_I = 0    # DC bus number (1 to 29997)
DC_BUS_TYPE: int = 1    # DC bus type
DC_PD = 2    # Pd, real power demand (MW)
DC_GS = 3    # Gs, shunt conductance (MW at V = 1.0 p.u.)
DC_BUS_AREA = 4    # area number, 1-100
DC_VM = 5    # Vm, voltage magnitude (p.u.)
DC_BASE_KV = 6    # baseKV, base voltage (kV)
DC_ZONE = 7   # zone, loss zone (1-999)
DC_VMAX = 8   # maxVm, maximum voltage magnitude (p.u.)
DC_VMIN = 9   # minVm, minimum voltage magnitude (p.u.)
DC_SL_FAC = 10   # scaling factor for special purposes

dc_bus_cols = 11



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import os

import numpy as np

from pandapower.__init__ import pp_dir
from pandapower.file_io import from_json
from pandapower.run import runpp
from pandapower.toolbox.grid_modification import select_subnet
from pandapower.topology.create_graph import create_nxgraph
from pandapower.topology.graph_searches import connected_components


def mv_oberrhein(scenario="load", cosphi_load=0.98, cosphi_pv=1.0, include_substations=False,
                 separation_by_sub=False, **kwargs):
    """
    Loads the Oberrhein network, a generic 20 kV network serviced by two 25 MVA HV/MV transformer
    stations. The network supplies 141 MV/LV substations and 6 MV loads through four MV feeders.
    The network layout is meshed, but the network is operated as a radial network with 6 open
    sectioning points.

    The network can be loaded with two different worst case scenarios for load and generation,
    which are defined by scaling factors for loads / generators as well as tap positions of the
    HV/MV transformers. These worst case scenarios are a good starting point for working with this
    network, but you are of course free to parametrize the network for your use case.

    The network also includes geographical information of lines and buses for plotting.

    OPTIONAL:
        **scenario** - (str, "load"): defines the scaling for load and generation

                - "load": high load scenario, load = 0.6 / sgen = 0, trafo taps [-2, -3]
                - "generation": high feed-in scenario: load = 0.1, generation = 0.8, trafo taps [0, 0]

        **cosphi_load** - (str, 0.98): cosine(phi) of the loads

        **cosphi_sgen** - (str, 1.0): cosine(phi) of the static generators

        **include_substations** - (bool, False): if True, the transformers of the MV/LV level are
        modelled, otherwise the loads representing the LV networks are connected directly to the
        MV node

        **separation_by_sub** - (bool, False): if True, the network gets separated into two
        sections, referring to both substations

    OUTPUT:
         **net** - pandapower network

         **net0, net1** - both sections of the pandapower network

    EXAMPLE:

        >>> from pandapower.networks.mv_oberrhein import mv_oberrhein
        >>> net = mv_oberrhein("generation")

        or with separation

        >>> net0, net1 = pandapower.networks.mv_oberrhein(separation_by_sub=True)
        """
    if include_substations:
        net = from_json(os.path.join(pp_dir, "networks", "mv_oberrhein_substations.json"), **kwargs)
        # geo.convert_epsg_bus_geodata(net, epsg_out=4326, epsg_in=31467)
        # geo.convert_geodata_to_geojson(net, lonlat=False)
    else:
        net = from_json(os.path.join(pp_dir, "networks", "mv_oberrhein.json"), **kwargs)
        # geo.convert_epsg_bus_geodata(net, epsg_out=4326, epsg_in=31467)
        # geo.convert_geodata_to_geojson(net, lonlat=False)
    net.load.q_mvar = np.tan(np.arccos(cosphi_load)) * net.load.p_mw
    net.sgen.q_mvar = np.tan(np.arccos(cosphi_pv)) * net.sgen.p_mw

    hv_trafos = net.trafo[net.trafo.sn_mva > 1].index
    net.trafo["tap_pos"] = net.trafo["tap_pos"].astype("float64")
    if scenario == "load":
        net.load.scaling = 0.6
        net.sgen.scaling = 0.0
        net.trafo.loc[hv_trafos, "tap_pos"] = [-2, -3]
    elif scenario == "generation":
        net.load.scaling = 0.1
        net.sgen.scaling = 0.8
        net.trafo.loc[hv_trafos, "tap_pos"] = [0, 0]
    else:
        raise ValueError("Unknown scenario %s - chose 'load' or 'generation'" % scenario)

    if separation_by_sub:
        # creating multigraph
        mg = create_nxgraph(net)
        # clustering connected buses
        zones = [list(area) for area in connected_components(mg)]
        net1 = select_subnet(net, buses=zones[0], include_switch_buses=False,
                             include_results=True, keep_everything_else=True)
        net0 = select_subnet(net, buses=zones[1], include_switch_buses=False,
                             include_results=True, keep_everything_else=True)

        runpp(net0)
        runpp(net1)
        net0.name = 'MV Oberrhein 0'
        net1.name = 'MV Oberrhein 1'
        # TODO: this should be added to the initial data not converted here.
        # geo.convert_geodata_to_geojson(net0)
        # geo.convert_geodata_to_geojson(net1)
        return net0, net1

    runpp(net)
    net.name = 'MV Oberrhein'
    # geo.convert_geodata_to_geojson(net)
    return net



import logging
import time

import pandas as pd

from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net
from pandapower.converter.cim.other_classes import Report, LogLevel, ReportCode

logger = logging.getLogger('cim.cim2pp.converter_classes.seriesCompensatorsCim16')

sc = cim_tools.get_pp_net_special_columns_dict()


class SeriesCompensatorsCim16:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def convert_series_compensators_cim16(self):
        time_start = time.time()
        self.logger.info("Start converting SeriesCompensators.")
        eq_sc = self._prepare_series_compensators_cim16()
        self.cimConverter.copy_to_pp('impedance', eq_sc)
        self.logger.info("Created %s impedance elements in %ss." % (eq_sc.index.size, time.time() - time_start))
        self.cimConverter.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
            message="Created %s impedance elements from SeriesCompensators in %ss." %
                    (eq_sc.index.size, time.time() - time_start)))

    def _prepare_series_compensators_cim16(self) -> pd.DataFrame:
        if 'sc' in self.cimConverter.cim.keys():
            ser_comp = self.cimConverter.merge_eq_sc_profile('SeriesCompensator')
        else:
            ser_comp = self.cimConverter.cim['eq']['SeriesCompensator']

        ser_comp = pd.merge(ser_comp,
                            pd.concat([self.cimConverter.cim['eq']['BaseVoltage'],
                                       self.cimConverter.cim['eq_bd']['BaseVoltage']],
                                      ignore_index=True)[['rdfId','nominalVoltage']].rename(
                                columns={'rdfId': 'BaseVoltage'}),
                            how='left', on='BaseVoltage')
        # fill the r21 and x21 values for impedance creation
        ser_comp['r21'] = ser_comp['r'].copy()
        ser_comp['x21'] = ser_comp['x'].copy()
        # set cim type
        ser_comp[sc['o_cl']] = 'SeriesCompensator'

        # add the buses
        eqs_length_before_merge = self.cimConverter.cim['eq']['SeriesCompensator'].index.size
        # until now self.cim['eq']['SeriesCompensator'] looks like:
        #   rdfId   name    r       ...
        #   _x01    bran1   0.056   ...
        #   _x02    bran2   0.471   ...
        # now join with the terminals and bus indexes
        ser_comp = pd.merge(ser_comp, self.cimConverter.bus_merge, how='left', on='rdfId')
        # now ser_comp looks like:
        #   rdfId   name    r       rdfId_Terminal  connected   ...
        #   _x01    bran1   0.056   termi025        True        ...
        #   _x01    bran1   0.056   termi223        True        ...
        #   _x02    bran2   0.471   termi154        True        ...
        #   _x02    bran2   0.471   termi199        True        ...
        # if each series compensator got two terminals, reduce back to one row to use fast slicing
        if ser_comp.index.size != eqs_length_before_merge * 2:
            self.logger.error("There is a problem at the SeriesCompensators source data: Not each SeriesCompensator "
                              "has two Terminals, %s Terminals should be given but there are %s Terminals available" %
                              (eqs_length_before_merge * 2, ser_comp.index.size))
            self.cimConverter.report_container.add_log(Report(
                level=LogLevel.ERROR, code=ReportCode.ERROR_CONVERTING,
                message="There is a problem at the SeriesCompensators source data: Not each SeriesCompensator "
                        "has two Terminals, %s Terminals should be given but there are %s Terminals available" %
                        (eqs_length_before_merge * 2, ser_comp.index.size)))
            dups = ser_comp.pivot_table(index=['rdfId'], aggfunc='size')
            dups = dups.loc[dups != 2]
            for rdfId, count in dups.items():
                self.logger.warning("The SeriesCompensator with RDF ID %s has %s Terminals!" % (rdfId, count))
                self.logger.warning("The SeriesCompensator data: \n%s" % ser_comp[ser_comp['rdfId'] == rdfId])
                self.cimConverter.report_container.add_log(Report(
                    level=LogLevel.ERROR, code=ReportCode.ERROR_CONVERTING,
                    message="The SeriesCompensator with RDF ID %s has %s Terminals!" % (rdfId, count)))
            ser_comp = ser_comp[0:0]
        # sort by RDF ID and the sequenceNumber to make sure r12 and r21 are in the correct order
        ser_comp = ser_comp.sort_values(by=['rdfId', 'sequenceNumber'])
        # copy the columns which are needed to reduce the ser_comp to one row per equivalent branch
        ser_comp['rdfId_Terminal2'] = ser_comp['rdfId_Terminal'].copy()
        ser_comp['connected2'] = ser_comp['connected'].copy()
        ser_comp['index_bus2'] = ser_comp['index_bus'].copy()
        ser_comp = ser_comp.reset_index()
        # here is where the magic happens: just remove the first value from the copied columns, reset the index
        # and replace the old column with the cut one. At least just remove the duplicates on column rdfId
        ser_comp.rdfId_Terminal2 = ser_comp.rdfId_Terminal2.iloc[1:].reset_index().rdfId_Terminal2
        ser_comp.connected2 = ser_comp.connected2.iloc[1:].reset_index().connected2
        ser_comp.index_bus2 = ser_comp.index_bus2.iloc[1:].reset_index().index_bus2
        ser_comp = ser_comp.drop_duplicates(['rdfId'], keep='first')
        if hasattr(self.cimConverter.net, 'sn_mva'):
            ser_comp['sn_mva'] = self.cimConverter.net['sn_mva']
        else:
            ser_comp['sn_mva'] = 1.
        # calculate z base in ohm
        ser_comp['z_base'] = ser_comp.nominalVoltage ** 2 / ser_comp.sn_mva
        ser_comp['rft_pu'] = ser_comp['r'] / ser_comp['z_base']
        ser_comp['xft_pu'] = ser_comp['x'] / ser_comp['z_base']
        ser_comp['rtf_pu'] = ser_comp['r21'] / ser_comp['z_base']
        ser_comp['xtf_pu'] = ser_comp['x21'] / ser_comp['z_base']
        ser_comp['rft0_pu'] = ser_comp['r0'] / ser_comp['z_base']
        ser_comp['xft0_pu'] = ser_comp['x0'] / ser_comp['z_base']
        ser_comp['rtf0_pu'] = ser_comp['r0'] / ser_comp['z_base']
        ser_comp['xtf0_pu'] = ser_comp['x0'] / ser_comp['z_base']
        ser_comp['gf_pu'] = 0.
        ser_comp['bf_pu'] = 0.
        ser_comp['gt_pu'] = 0.
        ser_comp['bt_pu'] = 0.
        ser_comp['in_service'] = ser_comp.connected & ser_comp.connected2
        ser_comp = ser_comp.rename(columns={'rdfId_Terminal': sc['t_from'], 'rdfId_Terminal2': sc['t_to'],
                                            'rdfId': sc['o_id'], 'index_bus': 'from_bus', 'index_bus2': 'to_bus'})
        return ser_comp



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

"""
Defines constants for named column indices to SSC matrix.

Some examples of usage, after defining the constants using the lines above,
are::

    ssc[3, SSC_STATUS] = 0                  # take SSC 4 out of service
    SSC_Q = ssc[:, SSC_Q]                   # extract reactive power vector

The index, name, and meaning of each column of the SSC matrix are given below:

columns 0-9 must be included in the input matrix (ssc in ppc dict)
    0.  C{SSC_BUS}                      SSC bus number
    1.  C{SSC_INTERNAL_BUS}             Internal bus number for SSC
    2.  C{SSC_R}                        SSC coupling transformer resistance (p.u.)
    3.  C{SSC_X}                        SSC coupling transformer reactance (p.u.)
    4.  C{SSC_SET_VM_PU}                SSC bus voltage magnitude set point (p.u.)
    5.  C{SSC_STATUS}                   Initial SSC status, 1 - in service, 0 - out of service
    6.  C{SSC_CONTROLLABLE}             Controllability flag for SSC
    7.  C{SSC_X_CONTROL_VA}             SSC internal voltage angle (radians)
    8.  C{SSC_X_CONTROL_VM}             SSC internal voltage magnitude (p.u.)
    9.  C{SSC_Q}                        reactive power absorbed (+) or injected (-) by SSC (MVAr)

"""

# define the indices
SSC_BUS = 0  # f, from bus number
SSC_INTERNAL_BUS = 1
SSC_R = 2  # (p.u.)
SSC_X = 3  # (p.u.)
SSC_SET_VM_PU = 4
SSC_STATUS = 5  # initial branch status, 1 - in service, 0 - out of service
SSC_CONTROLLABLE = 6
SSC_X_CONTROL_VA = 7 # va degrees ## check with roman
SSC_X_CONTROL_VM = 8 # (p.u)  vm
SSC_Q = 9 # result for Q

ssc_cols  = 10


import logging
import time
from typing import Dict, List

import pandas as pd

from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net
from pandapower.converter.cim.other_classes import Report, LogLevel, ReportCode

logger = logging.getLogger('cim.cim2pp.converter_classes.acLineSegmentsCim16')

sc = cim_tools.get_pp_net_special_columns_dict()


class DcLineSegmentsCim16:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def convert_dc_line_segments_cim16(self):
        time_start = time.time()
        self.logger.info("Start converting DCLineSegments.")
        eq_dc_line_segments = self._prepare_dc_line_segments_cim16()

        self.cimConverter.copy_to_pp('dcline', eq_dc_line_segments)

        self.logger.info("Created %s DC lines in %ss" % (eq_dc_line_segments.index.size, time.time() - time_start))
        self.cimConverter.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
            message="Created %s DC lines from DCLineSegments in %ss" %
                    (eq_dc_line_segments.index.size, time.time() - time_start)))

    def _prepare_dc_line_segments_cim16(self) -> pd.DataFrame:
        line_length_before_merge = self.cimConverter.cim['eq']['DCLineSegment'].index.size
        # until now self.cim['eq']['DCLineSegment'] looks like:
        #   rdfId   name    ...
        #   _x01    line1   ...
        #   _x02    line2   ...
        # now join with the terminals
        dc_line_segments = pd.merge(self.cimConverter.cim['eq']['DCLineSegment'], self.cimConverter.bus_merge,
                                    how='left', on='rdfId')
        dc_line_segments = dc_line_segments[['rdfId', 'name', 'description', 'ConnectivityNode', 'sequenceNumber']]
        dc_line_segments[sc['o_cl']] = 'DCLineSegment'
        # now dc_line_segments looks like:
        #   rdfId   name    rdfId_Terminal  connected   ...
        #   _x01    line1   termi025        True        ...
        #   _x01    line1   termi223        True        ...
        #   _x02    line2   termi154        True        ...
        #   _x02    line2   termi199        True        ...
        # if each switch got two terminals, reduce back to one line to use fast slicing
        if dc_line_segments.index.size != line_length_before_merge * 2:
            self.logger.error("Error processing the DCLineSegments, there is a problem with Terminals in the source "
                              "data!")
            self.cimConverter.report_container.add_log(Report(
                level=LogLevel.ERROR, code=ReportCode.ERROR_CONVERTING,
                message="Error processing the DCLineSegments, there is a problem with Terminals in the source data!"))
            return pd.DataFrame(None)
        dc_line_segments = dc_line_segments.reset_index()

        # now merge with the Converters
        converters = pd.merge(
            pd.concat([self.cimConverter.cim['eq']['CsConverter'], self.cimConverter.cim['eq']['VsConverter']],
                      ignore_index=True, sort=False),
            pd.concat([self.cimConverter.cim['ssh']['CsConverter'], self.cimConverter.cim['ssh']['VsConverter']],
                      ignore_index=True, sort=False),
            how='left', on='rdfId')
        if 'name' in converters.columns:
            converters = converters.drop(columns=['name'])
        # merge with the terminals
        converters = pd.merge(converters, self.cimConverter.bus_merge, how='left', on='rdfId')
        converters = converters.drop(columns=['sequenceNumber'])
        converters = converters.rename(columns={'rdfId': 'converters'})
        converter_terminals = pd.concat(
            [self.cimConverter.cim['eq']['Terminal'], self.cimConverter.cim['eq_bd']['Terminal']],
            ignore_index=True, sort=False)
        converter_terminals = converter_terminals[['rdfId']].rename(columns={'rdfId': 'rdfId_Terminal'})
        converters_t = pd.merge(converters, converter_terminals, how='inner', on='rdfId_Terminal')

        dc_line_segments = pd.merge(dc_line_segments, converters[['converters', 'ConnectivityNode']],
                                    how='left', on='ConnectivityNode')
        # get the missing converters (maybe there is a switch or something else between the line and the converter)
        t = self.cimConverter.cim['eq']['DCTerminal'][['DCNode', 'DCConductingEquipment', 'sequenceNumber']]
        t = t.rename(columns={'DCNode': 'ConnectivityNode', 'DCConductingEquipment': 'ConductingEquipment'})

        def search_converter(cn_ids: Dict[str, str], visited_cns: List[str]) -> str:
            new_cn_dict = dict()
            for one_cn, from_dev in cn_ids.items():
                # get the Terminals
                t_temp = t.loc[t['ConnectivityNode'] == one_cn, :]
                for _, one_t in t_temp.iterrows():
                    # prevent running backwards
                    if one_t['ConductingEquipment'] == from_dev:
                        continue
                    ids_temp = t.loc[(t['ConductingEquipment'] == one_t['ConductingEquipment']) & (
                            t['sequenceNumber'] != one_t['sequenceNumber'])]['ConnectivityNode'].values
                    for id_temp in ids_temp:
                        # check if the ConnectivityNode has a converter
                        if id_temp in converters['ConnectivityNode'].values:
                            # found the converter
                            return converters.loc[converters['ConnectivityNode'] == id_temp, 'converters'].values[0]
                        if id_temp not in visited_cns:
                            new_cn_dict[id_temp] = one_t['ConductingEquipment']
            if len(list(new_cn_dict.keys())) > 0:
                visited_cns.extend(list(cn_ids.keys()))
                return search_converter(cn_ids=new_cn_dict, visited_cns=visited_cns)

        for row_index, row in dc_line_segments[dc_line_segments['converters'].isna()].iterrows():
            conv = search_converter(cn_ids=dict({row['ConnectivityNode']: row['rdfId']}),
                                    visited_cns=[row['ConnectivityNode']])
            dc_line_segments.loc[row_index, 'converters'] = conv
            if conv is None:
                self.logger.warning("Problem with converting tht DC line %s: No ACDC converter found, maybe the DC "
                                    "part is too complex to reduce it to pandapower requirements!")
                self.cimConverter.report_container.add_log(Report(
                    level=LogLevel.WARNING, code=ReportCode.WARNING_CONVERTING,
                    message="Error processing the DCLineSegments, there is a problem with Terminals in the source "
                            "data!"))
        dc_line_segments = dc_line_segments.drop(columns=['ConnectivityNode'])
        dc_line_segments = pd.merge(dc_line_segments, converters_t, how='left', on='converters')
        dc_line_segments['targetUpcc'] = dc_line_segments['targetUpcc'].fillna(dc_line_segments['base_voltage_bus'])

        # copy the columns which are needed to reduce the dc_line_segments to one row per line
        dc_line_segments = dc_line_segments.sort_values(by=['rdfId', 'sequenceNumber'])
        # a list of DC line parameters which are used for each DC line end
        dc_line_segments = dc_line_segments.reset_index()
        copy_list = ['index_bus', 'rdfId_Terminal', 'connected', 'p', 'ratedUdc', 'targetUpcc', 'base_voltage_bus']
        for one_item in copy_list:
            # copy the columns which are required for each line end
            dc_line_segments[one_item + '2'] = dc_line_segments[one_item].copy()
            # cut the first element from the copied columns
            dc_line_segments[one_item + '2'] = dc_line_segments[one_item + '2'].iloc[1:].reset_index()[one_item + '2']
        del copy_list, one_item
        dc_line_segments = dc_line_segments.drop_duplicates(['rdfId'], keep='first')
        dc_line_segments = pd.merge(dc_line_segments,
                                    pd.DataFrame(dc_line_segments.pivot_table(index=['converters'], aggfunc='size'),
                                                 columns=['converter_dups']), how='left', on='converters')
        dc_line_segments['loss_mw'] = \
            abs(abs(dc_line_segments['p']) - abs(dc_line_segments['p2'])) / dc_line_segments['converter_dups']
        dc_line_segments['p_mw'] = dc_line_segments['p'] / dc_line_segments['converter_dups']
        dc_line_segments['loss_percent'] = 0
        dc_line_segments['vm_from_pu'] = dc_line_segments['targetUpcc'] / dc_line_segments['base_voltage_bus']
        dc_line_segments['vm_to_pu'] = dc_line_segments['targetUpcc2'] / dc_line_segments['base_voltage_bus2']
        if 'inService' not in dc_line_segments.columns:
            dc_line_segments['inService'] = True
        dc_line_segments['in_service'] = (dc_line_segments.connected & dc_line_segments.connected2
                                          & dc_line_segments.inService)
        dc_line_segments = dc_line_segments.rename(columns={
            'rdfId': sc['o_id'], 'rdfId_Terminal': sc['t_from'], 'rdfId_Terminal2': sc['t_to'], 'index_bus': 'from_bus',
            'index_bus2': 'to_bus'})

        return dc_line_segments



import sys
import os
import numpy as np
import tkinter as tk
from tkinter.filedialog import askdirectory
import pandas

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)
root_logger = logging.getLogger()

from pandapower.file_io import to_json, to_excel
from pandapower.toolbox.grid_modification import create_replacement_switch_for_branch, replace_xward_by_ward
from pandapower.diagnostic import diagnostic
from pandapower.converter.powerfactory.echo_off import echo_off, echo_on
from pandapower.converter.powerfactory.pp_import_functions import from_pf
from pandapower.converter.powerfactory import pf_export_functions as pef
from pandapower.converter.powerfactory import logger_setup as pflog
from pandapower.converter.powerfactory import gui
from pandapower.converter.powerfactory.validate import validate_pf_conversion

pandas.set_option('display.width', 1000)
pandas.set_option('display.max_columns', 100)


def browse_dst(input_panel, entry_path_dst):
    logger.debug('browsing directory')
    path = askdirectory(parent=input_panel, title='Provide path to the target directory '
                                                  'to save the grid in pandapower format')
    entry_path_dst.delete(0, tk.END)
    entry_path_dst.insert(tk.END, path)
    logger.debug('directory received: <%s>' % path)


def get_dst_dir(input_panel, entry_path_dst):
    logger.debug('getting destination path')
    dst_dir = entry_path_dst.get()

    if dst_dir == '':
        logger.warning('no destination directory given')
        browse_dst(input_panel, entry_path_dst)
        dst_dir = entry_path_dst.get()

    return dst_dir


def get_filename(entry_fname, save_as="JSON"):
    logger.debug('getting file name to save network')
    filename = entry_fname.get()
    if save_as == "JSON":
        goal_extension = ".json"
    elif save_as == "Excel":
        goal_extension = ".xlsx"
    else:
        raise NotImplementedError(f"extension {save_as} not implemented")
    _, extension = os.path.splitext(filename)
    logger.debug('filename extension: %s' % extension)
    if filename == '':
        raise RuntimeError('No filename given!')
    elif not extension == goal_extension:
        logger.debug('setting correct file extension')
        entry_fname.insert(len(filename), goal_extension)
        filename = entry_fname.get()
    return filename


def save_net(net, filepath, save_as):
    if save_as == "JSON":
        to_json(net, filepath)
    elif save_as == "Excel":
        to_excel(net, filepath)
    else:
        raise ValueError('tried to save grid as %s to %s and failed :(' % (save_as, filepath))


def exit_gracefully(app, input_panel, msg, is_err):
    if is_err:
        logger.error('Execution terminated: %s' % msg, exc_info=True)
    else:
        logger.info('Execution finished: %s' % msg)
    echo_on(app)
    input_panel.destroy()
    # del(app)
    # quit()
    sys.exit(1 if is_err else 0)


def run_export(app, pv_as_slack, pf_variable_p_loads, pf_variable_p_gen, scale_feeder_loads=False,
               flag_graphics='GPS', handle_us="Deactivate", save_as="JSON", tap_opt="nntap",
               export_controller=True, max_iter=None, create_sections=True):
    # gather objects from the project
    logger.info('gathering network elements')
    dict_net = pef.create_network_dict(app, flag_graphics)

    logger.info('running load flow calculation')
    echo_off(app, 1, 1, 1)
    pf_load_flow_failed = pef.run_load_flow(app, scale_feeder_loads=scale_feeder_loads)
    echo_off(app, 1, 1, 1)

    logger.info('starting import to pandapower')
    app.SetAttributeModeInternal(1)

    net = from_pf(dict_net, pv_as_slack=pv_as_slack, pf_variable_p_loads=pf_variable_p_loads,
                  pf_variable_p_gen=pf_variable_p_gen, flag_graphics=flag_graphics, tap_opt=tap_opt,
                  export_controller=export_controller, handle_us=handle_us, max_iter=max_iter,
                  create_sections=create_sections)
    # save a flag, whether the PowerFactory load flow failed
    app.SetAttributeModeInternal(0)
    net["pf_converged"] = not pf_load_flow_failed

    echo_on(app)
    return net


def run_verify(net, load_flow_params=None):
    logger.info('Validating import...')
    if load_flow_params is None:
        load_flow_params = {
            # 'tolerance_mva': 1e-9,  # tolerance of load flow calculation
            # 'calculate_voltage_angles': True,  # set True for meshed networks
            # 'init': 'dc',  # initialization of load flow: 'flat', 'dc', 'results'
            'PF_MAX_IT': 500  # Pypower option, maximal iterations, passed with kwargs
        }
    logger.debug('load flow params: %s' % load_flow_params)
    all_diffs = validate_pf_conversion(net, **load_flow_params)
    return all_diffs


def calc(app, input_panel, entry_path_dst, entry_fname, pv_as_slack, export_controller,
         replace_zero_branches, min_ohm_entry, replace_inf_branches, max_ohm_entry,
         is_to_verify, is_to_diagnostic, is_debug,
         pf_variable_p_loads, pf_variable_p_gen, flag_graphics, handle_us,
         save_as, tap_opt, max_iter_entry, create_sections_entry):
    # check if logger is to be in debug mode
    if is_debug():
        pflog.set_PF_level(root_logger, app_handler, 'DEBUG')
    else:
        pflog.set_PF_level(root_logger, app_handler, 'INFO')
    logger.debug('starting script')
    echo_off(app, err=1, warn=1, info=1)
    # start_button.config(state="disabled")
    # ask for destination path
    try:
        dst_dir = get_dst_dir(input_panel, entry_path_dst)
    except Exception as err:
        exit_gracefully(app, input_panel, err, True)
        return
    # ask for file name
    try:
        filename = get_filename(entry_fname, save_as())
    except RuntimeError as err:
        exit_gracefully(app, input_panel, err, True)
        return

    logger.info('the destination directory is: <%s>' % dst_dir)
    filepath = os.path.join(dst_dir, filename)

    # the actual export
    try:
        max_iter = int(max_iter_entry.get())
        logger.info("max_iter: %s" % max_iter)
        net = run_export(app, pv_as_slack(), pf_variable_p_loads(), pf_variable_p_gen(), scale_feeder_loads=False,
                         flag_graphics=flag_graphics(), handle_us=handle_us(), save_as=save_as(), tap_opt=tap_opt(),
                         export_controller=export_controller(), max_iter=max_iter, create_sections=create_sections_entry())
        if replace_zero_branches():
            min_ohm = float(min_ohm_entry.get())
            # to_replace = (np.abs(net.line.r_ohm_per_km * net.line.length_km +
            #                      1j * net.line.x_ohm_per_km * net.line.length_km) <= min_ohm) & net.line.in_service
            to_replace = (np.abs(net.line.x_ohm_per_km * net.line.length_km) <= min_ohm) & net.line.in_service

            for i in net.line.loc[to_replace].index.values:
                create_replacement_switch_for_branch(net, "line", i)
                net.line.at[i, "in_service"] = False
            if np.any(to_replace):
                logger.info(f"replaced {sum(to_replace)} lines with switches")

            # xward = net.xward[(np.abs(net.xward.r_ohm + 1j * net.xward.x_ohm) <= min_ohm) &
            #                   net.xward.in_service].index.values
            xward = net.xward[(np.abs(net.xward.x_ohm) <= min_ohm) & net.xward.in_service].index.values
            if len(xward) > 0:
                replace_xward_by_ward(net, index=xward, drop=False)
                logger.info(f"replaced {len(xward)} xwards with wards")

            zb_f_ohm = np.square(net.bus.loc[net.impedance.from_bus.values, "vn_kv"].values) / net.impedance.sn_mva
            zb_t_ohm = np.square(net.bus.loc[net.impedance.to_bus.values, "vn_kv"].values) / net.impedance.sn_mva
            # impedance = ((np.abs(net.impedance.rft_pu + 1j * net.impedance.xft_pu) <= min_ohm / zb_f_ohm) |
            #              (np.abs(net.impedance.rtf_pu + 1j * net.impedance.xtf_pu) <= min_ohm / zb_t_ohm) &
            #              net.impedance.in_service)
            impedance = ((np.abs(net.impedance.xft_pu) <= min_ohm / zb_f_ohm) |
                         (np.abs(net.impedance.xtf_pu) <= min_ohm / zb_t_ohm)) & net.impedance.in_service
            for i in net.impedance.loc[impedance].index.values:
                create_replacement_switch_for_branch(net, "impedance", i)
                net.impedance.at[i, "in_service"] = False
            if any(impedance):
                logger.info(f"replaced {sum(impedance)} impedance elements with switches")

            trafos = ((net.trafo.vk_percent/100 * np.square(net.trafo.vn_lv_kv) / net.trafo.sn_mva <= min_ohm) &
                      net.trafo.in_service)
            min_vk_percent = 4.
            net.trafo.loc[trafos, "vk_percent"] = min_vk_percent
            if any(trafos):
                logger.info(f"adjusted {sum(trafos)} 2W transformers by setting vk_percent to {min_vk_percent}")

            trafo3w = ((net.trafo3w.vk_hv_percent / 100 * np.square(net.trafo3w.vn_mv_kv) / net.trafo3w.sn_hv_mva <= min_ohm) |
                       (net.trafo3w.vk_mv_percent / 100 * np.square(net.trafo3w.vn_lv_kv) / net.trafo3w.sn_mv_mva <= min_ohm) |
                       (net.trafo3w.vk_lv_percent / 100 * np.square(net.trafo3w.vn_lv_kv) / net.trafo3w.sn_lv_mva <= min_ohm)) & net.trafo3w.in_service
            net.trafo3w.loc[trafo3w, ["vk_hv_percent", "vk_mv_percent", "vk_lv_percent"]] = min_vk_percent
            if any(trafo3w):
                logger.info(f"adjusted {sum(trafo3w)} 3W transformers by setting vk_percent to {min_vk_percent}")

        if replace_inf_branches():
            max_ohm = float(max_ohm_entry.get())
            to_replace = (np.abs(net.line.r_ohm_per_km * net.line.length_km +
                                 1j * net.line.x_ohm_per_km * net.line.length_km) >= max_ohm) & net.line.in_service

            if np.any(to_replace):
                net.line.loc[to_replace, "in_service"] = False
                logger.info(f"deactivated {sum(to_replace)} lines")

            xward = (np.abs(net.xward.r_ohm + 1j * net.xward.x_ohm) >= max_ohm) & net.xward.in_service
            if np.any(xward):
                net.xward.loc[xward, "in_service"] = False
                logger.info(f"deactivated {sum(xward)} xwards")

            zb_f_ohm = np.square(net.bus.loc[net.impedance.from_bus.values, "vn_kv"].values) / net.impedance.sn_mva
            zb_t_ohm = np.square(net.bus.loc[net.impedance.to_bus.values, "vn_kv"].values) / net.impedance.sn_mva
            impedance = ((np.abs(net.impedance.rft_pu + 1j * net.impedance.xft_pu) >= max_ohm / zb_f_ohm) |
                         (np.abs(net.impedance.rtf_pu + 1j * net.impedance.xtf_pu) >= max_ohm / zb_t_ohm)) & net.impedance.in_service
            if np.any(impedance):
                net.impedance.loc[impedance, "in_service"] = False
                logger.info(f"deactivated {sum(impedance)} impedance elements")

            max_vk_percent = 15
            trafos = (net.trafo.vk_percent/100 * np.square(net.trafo.vn_hv_kv) / net.trafo.sn_mva >= max_ohm) & net.trafo.in_service
            if np.any(trafos):
                # net.trafo.loc[trafos, "in_service"] = False
                net.trafo.loc[trafos, "vk_percent"] = max_vk_percent
                logger.info(f"adjusted {sum(trafos)} 2W transformers with vk_percent of {max_vk_percent}")

            trafo3w = ((net.trafo3w.vk_hv_percent / 100 * np.square(net.trafo3w.vn_hv_kv) / net.trafo3w.sn_hv_mva >= max_ohm) |
                       (net.trafo3w.vk_mv_percent / 100 * np.square(net.trafo3w.vn_mv_kv) / net.trafo3w.sn_mv_mva >= max_ohm) |
                       (net.trafo3w.vk_lv_percent / 100 * np.square(net.trafo3w.vn_hv_kv) / net.trafo3w.sn_lv_mva >= max_ohm)) & net.trafo3w.in_service
            if np.any(trafo3w):
                # net.trafo3w.loc[trafo3w, "in_service"] = False
                net.trafo3w.loc[trafo3w, "vk_hv_percent"] = np.fmin(max_vk_percent, net.trafo3w.loc[trafo3w, "vk_hv_percent"])
                net.trafo3w.loc[trafo3w, "vk_mv_percent"] = np.fmin(max_vk_percent, net.trafo3w.loc[trafo3w, "vk_mv_percent"])
                net.trafo3w.loc[trafo3w, "vk_lv_percent"] = np.fmin(max_vk_percent, net.trafo3w.loc[trafo3w, "vk_lv_percent"])
                logger.info(f"adjusted {sum(trafo3w)} 3W transformers with {max_vk_percent}")

        logger.info('saving file to: <%s>' % filepath)
        save_net(net, filepath, save_as())
        logger.info('exported net:\n %s' % net)
    except Exception as err:
        logger.error('Error while exporting net: %s', err, exc_info=True)
    else:
        if is_to_verify():
            try:
                run_verify(net)
            except Exception as err:
                logger.error('Error while verifying net: %s', err, exc_info=True)
            # logger.info('saving file to: <%s>' % filepath)
            # save_net(net, filepath, save_as())
            # logger.info('exported validated net')
        if is_to_diagnostic():
            try:
                diagnostic(net, report_style="compact", warnings_only=True, min_x_ohm=0.01)
            except Exception as err:
                logger.error('Error in diagnostic for net: %s', err, exc_info=True)
    root_logger.removeHandler(app_handler)
    exit_gracefully(app, input_panel, 'exiting script', False)


# if called from powerfactory, __name__ is also '__main__'
if __name__ == '__main__':
    try:
        import powerfactory as pf
    except:
        pass
    else:
        # power factory app
        app = pf.GetApplication()
        ### for debugging: uncomment app.Show() and add a breakpoint
        # app.Show()
        # echo_off(app)
        # logger, app_handler = pflog.setup_logger(app, __name__, 'INFO')
        # logger = logging.getLogger(__name__)
        app_handler = pflog.AppHandler(app, freeze_app_between_messages=True)
        root_logger.addHandler(app_handler)

        logger.info('starting application')
        # user to store the project and folders
        user = app.GetCurrentUser()
        logger.debug('got the current user <%s>' % user.loc_name)

        prj = app.GetActiveProject()
        logger.debug('asked for active project')

        if prj is not None:
            project_name = prj.GetAttribute('loc_name')
            logger.info('got the name of active project: <%s>' % project_name)
        elif False:  # user flag for debug
            app.Show()
            echo_on(app)
            app.ActivateProject('test')
            prj = app.GetActiveProject()
            project_name = prj.GetAttribute('loc_name')
            # net = calc('main', dst_dir=r'C:\pp_projects\test')
        else:
            logger.warning(
                'No project is activated. Please activate a project to perform PandaPower '
                'Export!')
            project_name = 'None'
            root_logger.removeHandler(app_handler)

        gui.make_gui(app, project_name, browse_dst, calc)



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.
from copy import deepcopy

from pandapower.control.util.auxiliary import get_controller_index
from pandapower.control.controller.trafo_control import TrafoController

try:
    import pandaplan.core.pplog as pplog
except:
    import logging as pplog

import warnings

logger = pplog.getLogger(__name__)


def control_diagnostic(net, respect_in_service=True):
    """
    Diagnostic function to find obvious mistakes in control data
    """
    # --- find and log same type controllers connected to same elements
    indices = list(net.controller.index)
    for idx in indices:
        current_controller = net.controller.object.loc[idx]
        parameters = deepcopy(current_controller.matching_params) if \
            "matching_params" in current_controller.__dict__ else {}
        if respect_in_service:
            if not net.controller.in_service.at[idx]:
                continue
            parameters["in_service"] = True
        same_type_existing_ctrl = get_controller_index(net, ctrl_type=type(current_controller),
                                                       parameters=parameters)
        if len(same_type_existing_ctrl) > 1:
            logger.info("Same type and same matching parameters controllers " + str([
                '%i' % i for i in same_type_existing_ctrl]) +
                        " could affect convergence.")
            for val in same_type_existing_ctrl:
                indices.remove(val)

    # --- find trafo controller of the same trafo
    trafo_ctrl = []
    for idx in net.controller.index:
        current_controller = net.controller.object.loc[idx]
        if issubclass(type(current_controller), TrafoController):
            trafo_ctrl += [idx]
    for idx in trafo_ctrl:
        current_controller = net.controller.object.loc[idx]
        parameters = {"element_index": current_controller.element_index,
                      "element": current_controller.element}
        if respect_in_service:
            if not net.controller.in_service.at[idx]:
                continue
            parameters["in_service"] = True
        trafo_ctrl_at_same_trafo = get_controller_index(net, parameters=parameters, idx=trafo_ctrl)
        if len(trafo_ctrl_at_same_trafo) > 1:
            logger.info(
                "Trafo Controllers %s at the %s transformer %s probably could affect convergence." %
                (str(['%i' % i for i in trafo_ctrl_at_same_trafo]), parameters['element'],
                 parameters["element_index"]))
            for val in trafo_ctrl_at_same_trafo:
                trafo_ctrl.remove(val)


def trafo_characteristic_table_diagnostic(net):
    logger.info("Checking transformer characteristic table")
    if "trafo_characteristic_table" not in net:
        logger.info("No transformer characteristic table found")
        return False
    cols2w = ["id_characteristic", "step", "voltage_ratio", "angle_deg", "vk_percent", "vkr_percent"]
    cols3w = ["id_characteristic", "step", "voltage_ratio", "angle_deg", "vk_hv_percent", "vkr_hv_percent",
              "vk_mv_percent", "vkr_mv_percent", "vk_lv_percent", "vkr_lv_percent"]
    warnings_count = 0
    for trafo_table, cols in zip(["trafo", "trafo3w"], [cols2w, cols3w]):
        if len(net[trafo_table]) == 0 or \
                not all(col in net[trafo_table] for col in ['id_characteristic_table', 'tap_dependency_table']) or \
                (not net[trafo_table]['id_characteristic_table'].notna().any() and
                 not net[trafo_table]['tap_dependency_table'].any()):
            logger.info("No %s with tap-dependent characteristics found." % trafo_table)
            continue
        # check if both tap_dependency_table & id_characteristic_table columns are populated
        mismatch_a = net[trafo_table][
            ((net[trafo_table]['tap_dependency_table']) & (net[trafo_table]['id_characteristic_table'].isna())) |
            ((~net[trafo_table]['tap_dependency_table']) & (net[trafo_table]['id_characteristic_table'].notna()))
            ].shape[0]
        if mismatch_a != 0:
            warnings.warn(f"{trafo_table}: found {mismatch_a} transformer(s) with not both "
                          f"tap_dependency_table and id_characteristic_table parameters populated. "
                          f"Power flow calculation will raise an error.", category=UserWarning)
            warnings_count += 1
        # check if both tap_dependency_table & tap_changer_type columns are populated
        mismatch_b = net[trafo_table][
            (net[trafo_table]['tap_dependency_table']) & (net[trafo_table]['tap_changer_type'].isna())
            ].shape[0]
        if mismatch_b != 0:
            warnings.warn(f"{trafo_table}: found {mismatch_b} transformer(s) with tap_dependency_table set to "
                          f"True and tap_changer_type parameter not populated. The characteristics from "
                          f"trafo_characteristic_table will not be considered.", category=UserWarning)
            warnings_count += 1
        # check if tap_changer_type is "Tabular" but tap_dependency_table is False
        mismatch_c = net[trafo_table][
            (~net[trafo_table]['tap_dependency_table']) & (net[trafo_table]['tap_changer_type'] == 'Tabular')
            ].shape[0]
        if mismatch_c != 0:
            warnings.warn(f"{trafo_table}: found {mismatch_c} transformer(s) with tap_changer_type parameter "
                          f"set to 'Tabular' but tap_dependency_table flag set to False. The characteristics from "
                          f"trafo_characteristic_table will not be considered.", category=UserWarning)
            warnings_count += 1
        # check if tap_changer_type is "Symmetrical" but tap_step_degree is not 90
        mismatch_d = net[trafo_table][
            (net[trafo_table]['tap_step_degree'] != 90) & (net[trafo_table]['tap_changer_type'] == 'Symmetrical')
            ].shape[0]
        if mismatch_d != 0:
            warnings.warn(f"{trafo_table}: found {mismatch_d} transformer(s) with tap_changer_type parameter "
                          f"set to 'Symmetrical' but tap_step_degree value not set to 90 degrees.",
                          category=UserWarning)
            warnings_count += 1
        # check if all relevant columns are populated in the trafo_characteristic_table
        temp = net[trafo_table].dropna(subset=["id_characteristic_table"])[
            ["tap_dependency_table", "id_characteristic_table"]]
        merged_df = temp.merge(net["trafo_characteristic_table"], left_on="id_characteristic_table",
                               right_on="id_characteristic", how="inner")
        unpopulated = merged_df.loc[~merged_df[cols].notna().all(axis=1)]
        if not unpopulated.empty:
            warnings.warn(f"There are some transformers in the {trafo_table} table with not all "
                          f"characteristics populated in the trafo_characteristic_table.", category=UserWarning)
            warnings_count += 1
        # check tap_dependency_table & id_characteristic_table column types
        if net[trafo_table]['tap_dependency_table'].dtype != 'bool':
            warnings.warn(f"The tap_dependency_table column in the {trafo_table} table is not of bool type.",
                          category=UserWarning)
            warnings_count += 1
        if net[trafo_table]['id_characteristic_table'].dtype != 'Int64':
            warnings.warn(f"The id_characteristic_table column in the {trafo_table} table is not of Int64 type.",
                          category=UserWarning)
            warnings_count += 1
        # check if all id_characteristic_table values are present in id_characteristic column
        # of trafo_characteristic_table
        if not net[trafo_table]['id_characteristic_table'].dropna().isin(
                net["trafo_characteristic_table"]['id_characteristic']).all():
            warnings.warn(f"Not all id_characteristic_table values in the {trafo_table} table are present "
                          f"in id_characteristic column of trafo_characteristic_table.", category=UserWarning)
            warnings_count += 1
    logger.info(f"{warnings_count} warnings were issued")
    return warnings_count == 0


def shunt_characteristic_table_diagnostic(net):
    logger.info("Checking shunt characteristic table")
    if "shunt_characteristic_table" not in net:
        logger.info("No shunt characteristic table found")
        return False
    cols = ["id_characteristic", "step", "q_mvar", "p_mw"]
    warnings_count = 0
    if len(net["shunt"]) == 0 or \
            not all(col in net["shunt"] for col in ['id_characteristic_table', 'step_dependency_table']) or \
            (not net["shunt"]['id_characteristic_table'].notna().any() and
             not net["shunt"]['step_dependency_table'].any()):
        logger.info("No shunt with step-dependent characteristics found.")
        return False
    # check if both step_dependency_table & id_characteristic_table columns are populated
    mismatch = net["shunt"][
        (net["shunt"]['step_dependency_table'] & net["shunt"]['id_characteristic_table'].isna()) |
        (~net["shunt"]['step_dependency_table'] & net["shunt"]['id_characteristic_table'].notna())
        ].shape[0]
    if mismatch != 0:
        warnings.warn(f"Found {mismatch} shunt(s) with not both "
                      f"step_dependency_table and id_characteristic_table parameters populated. "
                      f"Power flow calculation will raise an error.", category=UserWarning)
        warnings_count += 1
    # check if all relevant columns are populated in the shunt_characteristic_table
    temp = net["shunt"].dropna(subset=["id_characteristic_table"])[
        ["step_dependency_table", "id_characteristic_table"]]
    merged_df = temp.merge(net["shunt_characteristic_table"], left_on="id_characteristic_table",
                           right_on="id_characteristic", how="inner")
    unpopulated = merged_df.loc[~merged_df[cols].notna().all(axis=1)]
    if not unpopulated.empty:
        warnings.warn("There are some shunts with not all characteristics "
                      "populated in the shunt_characteristic_table.", category=UserWarning)
        warnings_count += 1
    # check step_dependency_table & id_characteristic_table column types
    if net["shunt"]['step_dependency_table'].dtype != 'bool':
        warnings.warn("The step_dependency_table column in the shunt table is not of bool type.",
                      category=UserWarning)
        warnings_count += 1
    if net["shunt"]['id_characteristic_table'].dtype != 'Int64':
        warnings.warn("The id_characteristic_table column in the shunt table is not of Int64 type.",
                      category=UserWarning)
        warnings_count += 1
    # check if all id_characteristic_table values are present in id_characteristic column
    # of shunt_characteristic_table
    if not net["shunt"]['id_characteristic_table'].dropna().isin(
            net["shunt_characteristic_table"]['id_characteristic']).all():
        warnings.warn("Not all id_characteristic_table values in the shunt table are present "
                      "in id_characteristic column of shunt_characteristic_table.", category=UserWarning)
        warnings_count += 1
    logger.info(f"{warnings_count} warnings were issued")
    return warnings_count == 0


def q_capability_curve_table_diagnostic(net, element):
    if element not in ["gen", "sgen"]:
        warnings.warn("The given element type is not valid for diagnostics. Please give gen or sgen "
                      "as a argument of the function", category=UserWarning)
        return False

    logger.info(f"Checking {element} Q capability curve characteristic table")
    if "q_capability_curve_table" not in net:
        logger.info("No Q capability curve table found")
        return False
    cols = ["id_q_capability_curve", "p_mw", "q_min_mvar", "q_max_mvar", "curve_style"]
    warnings_count = 0

    # Quick checks for element table and required columns
    if (len(net[element]) == 0 or not {"id_q_capability_curve_characteristic", "reactive_capability_curve", "curve_style"}.
            issubset(net[element].columns) or (not net[element]['id_q_capability_curve_characteristic'].notna().any()
            and not net[element]['reactive_capability_curve'].any()) and not net[element]['curve_style'].any()):
        logger.info(f"No {element} with Q capability curve table found.")
        return False

    # Check if both reactive_capability_curve & id_q_capability_curve_characteristic columns are populated
    mismatch = net[element][(net[element]['reactive_capability_curve'] & (
                (net[element]['id_q_capability_curve_characteristic'].isna()) | (
            net[element]['curve_style'].isna()))) | (~net[element]['reactive_capability_curve'] & (
                (net[element]['id_q_capability_curve_characteristic'].notna()) | (
            net[element]['curve_style'].notna())))].shape[0]
    if mismatch != 0:
        warnings.warn(f"Found {mismatch} {element}(s) with mismatched between curve_style, "
                      f"reactive_capability_curve and id_q_capability_curve_characteristic parameters populated. "
                      f"Power flow calculation will raise an error.", category=UserWarning)
        warnings_count += 1

    # Validate relevant columns in q_capability_curve_table
    temp = net[element].dropna(subset=["id_q_capability_curve_characteristic"])[
        ["reactive_capability_curve", "id_q_capability_curve_characteristic", "curve_style"]]
    merged_df = temp.merge(net["q_capability_curve_table"], left_on="id_q_capability_curve_characteristic",
                           right_on="id_q_capability_curve", how="inner")

    if not merged_df[cols].notna().all(axis=1).all():
        warnings.warn(f"There are some {element}(s) with not all characteristics "
                      "populated in the q_capability_curve_table.", category=UserWarning)
        warnings_count += 1

    # Check reactive_capability_curve & id_characteristic_table column types
    if net[element]['reactive_capability_curve'].dtype != 'bool':
        warnings.warn(f"The reactive_capability_curve column in the {element} table is not of bool type.",
                      category=UserWarning)
        warnings_count += 1

    if net[element]['id_q_capability_curve_characteristic'].dtype != 'Int64':
        warnings.warn(f"The id_characteristic_table column in the {element} table is not of Int64 type.",
                      category=UserWarning)
        warnings_count += 1

    # check the curve style is known or not
    curve_df = net[element]['curve_style']
    curve_df = curve_df[~curve_df.isin(["straightLineYValues", "constantYValue"])].dropna()
    if curve_df.count() > 0:
        warnings.warn(f"There are {curve_df.count()} unknown curve style in curve_style column of the "
                      f"{element} table", category=UserWarning)
        warnings_count += 1

    # check if all id_q_capability_curve_characteristic values are present in id_q_capability_curve column
    # of q_capability_curve_table
    if not net[element]['id_q_capability_curve_characteristic'].dropna().isin(
            net["q_capability_curve_table"]['id_q_capability_curve']).any():
        warnings.warn(f"Not all id_q_capability_curve_characteristic values of {element} are present in the "
                      f"q_capability_curve_table.", category=UserWarning)
        warnings_count += 1

    logger.info(f"{warnings_count} warnings were issued")
    return warnings_count == 0



# -*- coding: utf-8 -*-

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


from numpy import complex128, float64, int64, r_, zeros, empty, array
from scipy.sparse import csr_matrix as sparse, vstack, hstack

from pandapower.pypower.dSbus_dV import dSbus_dV

try:
    # numba functions
    from pandapower.pf.create_jacobian_numba import create_J, create_J2, create_J_ds
    from pandapower.pf.dSbus_dV_numba import dSbus_dV_numba_sparse
except ImportError:
    pass


def _create_J_with_numba(Ybus, V, refpvpq, pvpq, pq, createJ, pvpq_lookup, nref, npv, npq, slack_weights, dist_slack):
    Ibus = zeros(len(V), dtype=complex128)
    # create Jacobian from fast calc of dS_dV
    dVm_x, dVa_x = dSbus_dV_numba_sparse(Ybus.data, Ybus.indptr, Ybus.indices, V, V / abs(V), Ibus)
    # data in J, space preallocated is bigger than acutal Jx -> will be reduced later on
    Jx = empty(len(dVm_x) * 4, dtype=float64)
    # row pointer, dimension = pvpq.shape[0] + pq.shape[0] + 1
    if dist_slack:
        Jp = zeros(refpvpq.shape[0] + pq.shape[0] + 1, dtype=int64)
    else:
        Jp = zeros(pvpq.shape[0] + pq.shape[0] + 1, dtype=int64)
    # indices, same with the preallocated space (see Jx)
    Jj = empty(len(dVm_x) * 4, dtype=int64)

    # fill Jx, Jj and Jp
    createJ(dVm_x, dVa_x, Ybus.indptr, Ybus.indices, pvpq_lookup, refpvpq, pvpq, pq, Jx, Jj, Jp, slack_weights)

    # resize before generating the scipy sparse matrix
    Jx.resize(Jp[-1], refcheck=False)
    Jj.resize(Jp[-1], refcheck=False)

    # todo: why not replace npv by pv.shape[0] etc.?
    # generate scipy sparse matrix
    if dist_slack:
        dimJ = nref + npv + npq + npq
    else:
        dimJ = npv + npq + npq
    J = sparse((Jx, Jj, Jp), shape=(dimJ, dimJ))

    return J


def _create_J_without_numba(Ybus, V, ref, pvpq, pq, slack_weights, dist_slack):
    # create Jacobian with standard pypower implementation.
    dS_dVm, dS_dVa = dSbus_dV(Ybus, V)

    ## evaluate Jacobian

    if dist_slack:
        rows_pvpq = array(r_[ref, pvpq]).T
        cols_pvpq = r_[ref[1:], pvpq]
        J11 = dS_dVa[rows_pvpq, :][:, cols_pvpq].real
        J12 = dS_dVm[rows_pvpq, :][:, pq].real
    else:
        rows_pvpq = array([pvpq]).T
        cols_pvpq = pvpq
        J11 = dS_dVa[rows_pvpq, cols_pvpq].real
        J12 = dS_dVm[rows_pvpq, pq].real
    if len(pq) > 0 or dist_slack:
        J21 = dS_dVa[array([pq]).T, cols_pvpq].imag
        J22 = dS_dVm[array([pq]).T, pq].imag
        if dist_slack:
            J10 = sparse(slack_weights[rows_pvpq].reshape(-1,1))
            J20 = sparse(zeros(shape=(len(pq), 1)))
            J = vstack([
                hstack([J10, J11, J12]),
                hstack([J20, J21, J22])
            ], format="csr")
        else:
            J = vstack([
                hstack([J11, J12]),
                hstack([J21, J22])
            ], format="csr")
    else:
        J = vstack([
            hstack([J11, J12])
        ], format="csr")
    return J


def create_jacobian_matrix(Ybus, V, ref, refpvpq, pvpq, pq, createJ, pvpq_lookup, nref, npv, npq, numba, slack_weights, dist_slack):
    if numba:
        J = _create_J_with_numba(Ybus, V, refpvpq, pvpq, pq, createJ, pvpq_lookup, nref, npv, npq, slack_weights, dist_slack)
    else:
        J = _create_J_without_numba(Ybus, V, ref, pvpq, pq, slack_weights, dist_slack)
    return J


def get_fastest_jacobian_function(pvpq, pq, numba, dist_slack):
    if numba:
        if dist_slack:
            create_jacobian = create_J_ds
        elif len(pvpq) == len(pq):
            create_jacobian = create_J2
        else:
            create_jacobian = create_J
    else:
        create_jacobian = None
    return create_jacobian



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


"""Defines constants for named column indices to TCSC matrix.

Some examples of usage, after defining the constants using the line above,
are:

    tcsc[3, TCSC_STATUS] = 0              # take TCSC 4 out of service
    Pset = tcsc[:, TCSC_SET_P]            # extract TCSC power set point vector

The index, name and meaning of each column of the TCSC matrix is given
below:

columns 0-9 must be included in input matrix (tcsc in ppc dict)
    0.  C{TCSC_F_BUS}                       TCSC from bus number
    1.  C{TCSC_T_BUS}                       TCSC to bus number
    2.  C{TCSC_X_L}                         inductive reactance (p.u.)
    3.  C{TCSC_X_CVAR}                      capacitive reactance (p.u.)
    4.  C{TCSC_SET_P}                       TCSC active power set point (p.u.)
    5.  C{TCSC_THYRISTOR_FIRING_ANGLE}      TCSC thyristor firing angle (radians)
    6.  C{TCSC_STATUS}                      initial TCSC status, 1 - in service, 0 - out of service
    7.  C{TCSC_CONTROLLABLE}                controllability flag for TCSC
    8.  C{TCSC_MIN_FIRING_ANGLE}            minimum thyristor firing angle (radians)
    9.  C{TCSC_MAX_FIRING_ANGLE}            maximum thyristor firing angle (radians)

columns 10-16 are filled with values after power flow

    10. C{PF}                               real power injected at "from" bus end (MW)
    11. C{QF}                               reactive power injected at "from" bus end (MVAr)
    12. C{PT}                               real power injected at "to" bus end (MW)
    13. C{QT}                               reactive power injected at "to" bus end (MVAr)
    14. C{IF}                               current injected at "from" bus end (kA)
    15. C{IT}                               current injected at "to" bus end (kA)
    16. C{TCSC_X_PU}                        reactance in per unit (p.u.)

"""

# define the indices
TCSC_F_BUS = 0  # f, from bus number
TCSC_T_BUS = 1  # t, to bus number
TCSC_X_L = 2  # (p.u.)
TCSC_X_CVAR = 3  # (p.u.)
TCSC_SET_P = 4  # (p.u.)
TCSC_THYRISTOR_FIRING_ANGLE = 5
TCSC_STATUS = 6  # initial branch status, 1 - in service, 0 - out of service
TCSC_CONTROLLABLE = 7
TCSC_MIN_FIRING_ANGLE = 8
TCSC_MAX_FIRING_ANGLE = 9

TCSC_PF = 10
TCSC_QF = 11
TCSC_PT = 12
TCSC_QT = 13

TCSC_IF = 14
TCSC_IT = 15
TCSC_X_PU = 16

tcsc_cols = 17



import pandapower.control.basic_controller
import pandapower.control.controller
from pandapower.control.controller.DERController.der_control import DERController
from pandapower.control.controller.characteristic_control import CharacteristicControl
# --- Controller ---
from pandapower.control.controller.const_control import ConstControl
from pandapower.control.controller.pq_control import PQController
from pandapower.control.controller.shunt_control import DiscreteShuntController
from pandapower.control.controller.station_control import BinarySearchControl, DroopControl
from pandapower.control.controller.trafo.ContinuousTapControl import ContinuousTapControl
from pandapower.control.controller.trafo.DiscreteTapControl import DiscreteTapControl
from pandapower.control.controller.trafo.TapDependentImpedance import TapDependentImpedance
from pandapower.control.controller.trafo.VmSetTapControl import VmSetTapControl
from pandapower.control.controller.trafo_control import TrafoController
# --- Other ---
from pandapower.control.run_control import *
from pandapower.control.util.auxiliary import (
    get_controller_index,
    plot_characteristic,
    _create_trafo_characteristics,
    create_trafo_characteristic_object,
    create_q_capability_curve_characteristics_object
)
from pandapower.control.util.characteristic import Characteristic, SplineCharacteristic
from pandapower.control.util.diagnostic import control_diagnostic, trafo_characteristic_table_diagnostic


# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.
import warnings

import numpy as np
import pandas as pd

from pandapower.auxiliary import _sum_by_group
from pandapower.pypower.idx_bus import BASE_KV, VM, VA
from pandapower.pypower.idx_brch import TAP
from pandapower.pypower.idx_brch_sc import IKSS_F, IKSS_T, IP_F, IP_T, ITH_F, ITH_T, \
    PKSS_F, QKSS_F, PKSS_T, QKSS_T, VKSS_MAGN_F, VKSS_MAGN_T, VKSS_ANGLE_F, VKSS_ANGLE_T, IKSS_ANGLE_F, IKSS_ANGLE_T
from pandapower.pypower.idx_bus_sc import C_MIN, C_MAX, KAPPA, R_EQUIV, IKSS1, IP, ITH, \
    X_EQUIV, IKSS2, IKCV, M, R_EQUIV_OHM, X_EQUIV_OHM, V_G, K_SG, SKSS, \
    PHI_IKSS1_DEGREE, PHI_IKSS2_DEGREE, PHI_IKCV_DEGREE
from pandapower.shortcircuit.impedance import _calc_zbus_diag

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def _calc_ikss(net, ppci, bus_idx):
    fault = net._options["fault"]
    case = net._options["case"]
    c = ppci["bus"][bus_idx, C_MIN] if case == "min" else ppci["bus"][bus_idx, C_MAX]
    baseI = ppci["internal"]["baseI"] = ppci["bus"][:, BASE_KV] * np.sqrt(3) / ppci["baseMVA"]

    # Only for test, should correspondant to PF result
    baseZ = ppci["bus"][bus_idx, BASE_KV] ** 2 / ppci["baseMVA"]
    ppci["bus"][bus_idx, R_EQUIV_OHM] = baseZ * ppci["bus"][bus_idx, R_EQUIV]
    ppci["bus"][bus_idx, X_EQUIV_OHM] = baseZ * ppci["bus"][bus_idx, X_EQUIV]

    # init V0 for the superposition method
    # todo: explain the background better
    n_sc_bus = np.shape(bus_idx)[0]
    n_bus = ppci["bus"].shape[0]

    valid_V = True
    if net._options.get("use_pre_fault_voltage", False):
        V0 = np.full((n_bus, n_sc_bus), ppci["bus"][:, [VM]] * np.exp(np.deg2rad(ppci["bus"][:, [VA]]) * 1j))
    else:
        V0 = np.full((n_bus, n_sc_bus), ppci["bus"][bus_idx, C_MAX if case == "max" else C_MIN], dtype=np.complex128)

        if np.any(ppci["branch"][:, TAP] != 1):
            msg = "Calculation does not support calculation of voltages and branch powers for grids that have" \
                  " transformers with rated voltages unequal to bus voltages. Change the transformer data or" \
                  "try using the superposition method by passing 'use_pre_fault_voltage=True'"
            # raise NotImplementedError(msg)
            warnings.warn(msg)
            valid_V = False
    ppci["internal"]["valid_V"] = valid_V

    z_equiv = ppci["bus"][bus_idx, R_EQUIV] + ppci["bus"][bus_idx, X_EQUIV] * 1j  # removed the abs()
    ikss1 = V0[bus_idx, np.arange(n_sc_bus)] / z_equiv

    if fault == "3ph":
        if net["_options"]["inverse_y"]:
            Zbus = ppci["internal"]["Zbus"]
            # I don't know how to do this without reshape
            V_ikss = V0 - ikss1 * Zbus[:, bus_idx] if valid_V else -ikss1 * Zbus[:, bus_idx]
        else:
            ybus_fact = ppci["internal"]["ybus_fact"]
            V_ikss = np.zeros((n_bus, n_sc_bus), dtype=np.complex128)
            for ix, b in enumerate(bus_idx):
                ikss = np.zeros((n_bus, 1), dtype=np.complex128)
                ikss[b] = ikss1[ix]
                V_ikss[:, [ix]] = V0[:, [ix]] - ybus_fact(ikss) if valid_V else -ybus_fact(ikss)

        V_ikss[np.abs(V_ikss) < 1e-10] = 0
        # ikss1 = -Ybus.dot(V_ikss) / baseI.reshape(-1, 1)
        # ikss1 = ikss1[bus_idx]
        # ikss1 = c / z_equiv / ppci["bus"][bus_idx, BASE_KV] / np.sqrt(3) * ppci["baseMVA"]
        # ikss1 = c / z_equiv / baseI[bus_idx]  # should be same as above
        ikss1 /= baseI[bus_idx]
        # added abs here:
        ppci["bus"][bus_idx, IKSS1] = abs(ikss1)
        # added angle calculation in degree:
        ppci["bus"][bus_idx, PHI_IKSS1_DEGREE] = np.angle(ikss1, deg=True)
        ppci["internal"]["V_ikss"] = V_ikss
    elif fault == "2ph":
        ppci["bus"][bus_idx, IKSS1] = np.abs(c / z_equiv / ppci["bus"][bus_idx, BASE_KV] / 2 * ppci["baseMVA"])

    _current_source_current(net, ppci, bus_idx)

    ikss = ppci["bus"][bus_idx, IKSS1] + ppci["bus"][bus_idx, IKSS2]
    if fault == "3ph":
        ppci["bus"][bus_idx, SKSS] = np.sqrt(3) * ikss * ppci["bus"][bus_idx, BASE_KV]
    elif fault == "2ph":
        ppci["bus"][bus_idx, SKSS] = ikss * ppci["bus"][bus_idx, BASE_KV] / np.sqrt(3)

    # Correct voltage of generator bus inside power station
    if np.any(~np.isnan(ppci["bus"][:, K_SG])):
        gen_bus_idx = bus_idx[~np.isnan(ppci["bus"][bus_idx, K_SG])]
        ppci["bus"][gen_bus_idx, IKSS1] *=\
            (ppci["bus"][gen_bus_idx, V_G] / ppci["bus"][gen_bus_idx, BASE_KV])
        ppci["bus"][gen_bus_idx, SKSS] *=\
            (ppci["bus"][gen_bus_idx, V_G] / ppci["bus"][gen_bus_idx, BASE_KV])


def _calc_ikss_1ph(net, ppci, ppci_0, bus_idx):
    case = net._options["case"]
    c = ppci["bus"][bus_idx, C_MIN] if case == "min" else ppci["bus"][bus_idx, C_MAX]
    ppci["internal"]["baseI"] = ppci["bus"][:, BASE_KV] * np.sqrt(3) / ppci["baseMVA"]
    ppci_0["internal"]["baseI"] = ppci_0["bus"][:, BASE_KV] * np.sqrt(3) / ppci_0["baseMVA"]

    z_equiv = abs((ppci["bus"][bus_idx, R_EQUIV] + ppci["bus"][bus_idx, X_EQUIV] * 1j) * 2 +
                  (ppci_0["bus"][bus_idx, R_EQUIV] + ppci_0["bus"][bus_idx, X_EQUIV] * 1j))

    # Only for test, should correspondant to PF result
    baseZ = ppci["bus"][bus_idx, BASE_KV] ** 2 / ppci["baseMVA"]
    ppci["bus"][bus_idx, R_EQUIV_OHM] = baseZ * ppci['bus'][bus_idx, R_EQUIV]
    ppci["bus"][bus_idx, X_EQUIV_OHM] = baseZ * ppci['bus'][bus_idx, X_EQUIV]
    ppci_0["bus"][bus_idx, R_EQUIV_OHM] = baseZ * ppci_0['bus'][bus_idx, R_EQUIV]
    ppci_0["bus"][bus_idx, X_EQUIV_OHM] = baseZ * ppci_0['bus'][bus_idx, X_EQUIV]

    # # ppci["bus"][bus_idx, IKSS1] = abs(c * ppci["internal"]["baseI"][bus_idx] * ppci["baseMVA"] / (z_equiv * baseZ))
    # # ppci_0["bus"][bus_idx, IKSS1] = abs(c * ppci_0["internal"]["baseI"][bus_idx] * ppci["baseMVA"] / (z_equiv * baseZ))
    # ppci["bus"][bus_idx, IKSS1] = abs(np.sqrt(3) * c / z_equiv / ppci["bus"][bus_idx, BASE_KV] * ppci["baseMVA"])
    # ppci_0["bus"][bus_idx, IKSS1] = abs(np.sqrt(3) * c / z_equiv / ppci_0["bus"][bus_idx, BASE_KV] * ppci["baseMVA"])
    ppci["bus"][bus_idx, IKSS1] = np.sqrt(3) * c / z_equiv / ppci["bus"][bus_idx, BASE_KV] * ppci["baseMVA"]
    ppci_0["bus"][bus_idx, IKSS1] = np.sqrt(3) * c / z_equiv / ppci_0["bus"][bus_idx, BASE_KV] * ppci_0["baseMVA"]
    ppci["bus"][bus_idx, PHI_IKSS1_DEGREE] = 0
    ppci_0["bus"][bus_idx, PHI_IKSS1_DEGREE] = 0

    _current_source_current(net, ppci, bus_idx)


def _current_source_current(net, ppci, bus_idx):
    case = net._options["case"]
    fault_impedance = net._options["fault_impedance"]
    ppci["bus"][:, IKCV] = 0
    ppci["bus"][:, PHI_IKCV_DEGREE] = -90
    ppci["bus"][:, IKSS2] = 0
    type_c = net._options["use_pre_fault_voltage"]
    # sgen current source contribution only for Type A and case "max" or type C:
    if case != "max" and not type_c:
        return

    bus_lookup = net["_pd2ppc_lookups"]["bus"]
    fault = net._options["fault"]
    # _is_elements_final exists for some reason, and weirdly it can be different than _is_elements. 
    # it is not documented anywhere why it exists and I don't have any time to find out, but this here fixes the problem.

    if np.all(net.sgen.current_source.values):
        sgen = net.sgen[net._is_elements_final["sgen"]]
    else:
        sgen = net.sgen[net._is_elements_final["sgen"] & net.sgen.current_source]
    if len(sgen) == 0:
        return
    if any(pd.isnull(sgen.sn_mva)):
        raise ValueError("sn_mva needs to be specified for all sgens in net.sgen.sn_mva")
    if "current_angle_degree" in sgen.columns:
        sgen_angle = np.deg2rad(sgen.current_angle_degree.values)
    else:
        sgen_angle = None

    baseI = ppci["internal"]["baseI"]
    sgen_buses = sgen.bus.values
    sgen_buses_ppc = bus_lookup[sgen_buses]

    if "k" not in sgen:
        raise ValueError("Nominal to short-circuit current has to specified in net.sgen.k")
    if type_c and "kappa" not in sgen:
        raise ValueError("Max. short-circuit current in p.u. must be specified in net.sgen.kappa for the "
                         "short-circuit calculation with superposition method")

    if type_c:
        # voltage difference between pre-fault condition and the voltage from the calculation with
        # rotating machines only is used to determine the current injection. The parameter kappa here
        # is used to denote the maximal current contribution of the sgen in the short-circuit case "Type C"
        V_ikss = ppci["internal"]["V_ikss"]
        i_sgen_n_pu = (sgen.sn_mva.values.reshape(-1, 1) / net.sn_mva)
        delta_V = ppci["bus"][sgen_buses_ppc][:, [VM]] - np.abs(V_ikss[sgen_buses_ppc])
        i_sgen_pu = np.where(sgen.k.values.reshape(-1, 1) * delta_V < sgen.kappa.values.reshape(-1, 1),
                             sgen.k.values.reshape(-1, 1) * i_sgen_n_pu * delta_V,
                             i_sgen_n_pu * sgen.kappa.values.reshape(-1, 1))
        i_sgen_pu = np.abs(i_sgen_pu)
        extra_angle = ppci["bus"][sgen_buses_ppc, VA]
    else:
        i_sgen_pu = (sgen.sn_mva.values / net.sn_mva * sgen.k.values)
        extra_angle = 0

    if sgen_angle is not None and fault == "3ph":
        i_sgen_pu = i_sgen_pu * np.exp(sgen_angle * 1j)
    # if case == "min":
    #     i_sgen_pu *= 0

    buses, ikcv_pu, _ = _sum_by_group(sgen_buses_ppc, i_sgen_pu, i_sgen_pu)
    ikcv_pu = ikcv_pu.flatten()
    ppci["bus"][buses, [IKCV]] = ikcv_pu if sgen_angle is None else np.abs(ikcv_pu)
    if sgen_angle is not None and fault == "3ph":
        ppci["bus"][buses, PHI_IKCV_DEGREE] = np.angle(ikcv_pu, deg=True)

    if net["_options"]["inverse_y"]:
        Zbus = ppci["internal"]["Zbus"]
        diagZ = np.diag(Zbus).copy()  # here diagZ is not writeable
        if sgen_angle is None and fault == "3ph":
            ppci["bus"][buses, PHI_IKCV_DEGREE] = -np.angle(diagZ[buses], deg=True) + extra_angle
        diagZ[bus_idx] += fault_impedance
        i_kss_2 = 1 / diagZ * np.dot(Zbus, ppci["bus"][:, IKCV] * np.exp(np.deg2rad(ppci["bus"][:, PHI_IKCV_DEGREE]) * 1j))
    else:
        ybus_fact = ppci["internal"]["ybus_fact"]
        diagZ = _calc_zbus_diag(net, ppci)
        if sgen_angle is None and fault == "3ph":
            ppci["bus"][buses, PHI_IKCV_DEGREE] = -np.angle(diagZ[buses], deg=True) + extra_angle
        diagZ[bus_idx] += fault_impedance
        i_kss_2 = ybus_fact(ppci["bus"][:, IKCV] * np.exp(np.deg2rad(ppci["bus"][:, PHI_IKCV_DEGREE]) * 1j)) / diagZ

    ppci["bus"][:, IKSS2] = np.abs(i_kss_2 / baseI)
    ppci["bus"][:, PHI_IKSS2_DEGREE] = np.angle(i_kss_2, deg=True) if fault == "3ph" else 0
    ppci["bus"][buses, IKCV] /= baseI[buses]


def _calc_ip(net, ppci):
    ip = np.sqrt(2) * (ppci["bus"][:, KAPPA] * ppci["bus"][:, IKSS1] + ppci["bus"][:, IKSS2])
    ppci["bus"][:, IP] = ip


def _calc_ith(net, ppci):
    tk_s = net["_options"]["tk_s"]
    kappa = ppci["bus"][:, KAPPA]
    f = 50
    n = 1
    m = (np.exp(4 * f * tk_s * np.log(kappa - 1)) - 1) / (2 * f * tk_s * np.log(kappa - 1))
    m[np.where(kappa > 1.99)] = 0
    ppci["bus"][:, M] = m
    ith = (ppci["bus"][:, IKSS1] + ppci["bus"][:, IKSS2]) * np.sqrt(m + n)
    ppci["bus"][:, ITH] = ith


# TODO: Ib for generation close bus
# def _calc_ib_generator(net, ppci):
#     # Zbus = ppci["internal"]["Zbus"]
#     # baseI = ppci["internal"]["baseI"]
#     tk_s = net._options['tk_s']
#     c = 1.1

#     z_equiv = ppci["bus"][:, R_EQUIV] + ppci["bus"][:, X_EQUIV] * 1j
#     I_ikss = c / z_equiv / ppci["bus"][:, BASE_KV] / np.sqrt(3) * ppci["baseMVA"]

#     # calculate voltage source branch current
#     # I_ikss = ppci["bus"][:, IKSS1]
#     # V_ikss = (I_ikss * baseI) * Zbus

#     gen = net["gen"][net._is_elements["gen"]]
#     gen_vn_kv = gen.vn_kv.values

#     # Check difference ext_grid and gen
#     gen_buses = ppci['gen'][:, GEN_BUS].astype(np.int64)
#     gen_mbase = ppci['gen'][:, MBASE]
#     gen_i_rg = gen_mbase / (np.sqrt(3) * gen_vn_kv)

#     gen_buses_ppc, gen_sn_mva, I_rG = _sum_by_group(gen_buses, gen_mbase, gen_i_rg)

#     # shunt admittance of generator buses and generator short circuit current
#     # YS = ppci["bus"][gen_buses_ppc, GS] + ppci["bus"][gen_buses_ppc, BS] * 1j
#     # I_kG = V_ikss.T[:, gen_buses_ppc] * YS / baseI[gen_buses_ppc]

#     xdss_pu = gen.xdss_pu.values
#     rdss_pu = gen.rdss_pu.values
#     cosphi = gen.cos_phi.values
#     X_dsss = xdss_pu * np.square(gen_vn_kv) / gen_mbase
#     R_dsss = rdss_pu * np.square(gen_vn_kv) / gen_mbase

#     K_G = ppci['bus'][gen_buses, BASE_KV] / gen_vn_kv * c / (1 + xdss_pu * np.sin(np.arccos(cosphi)))
#     Z_G = (R_dsss + 1j * X_dsss)

#     I_kG = c * ppci['bus'][gen_buses, BASE_KV] / np.sqrt(3) / (Z_G * K_G) * ppci["baseMVA"]

#     dV_G = 1j * X_dsss * K_G * I_kG
#     V_Is = c * ppci['bus'][gen_buses, BASE_KV] / np.sqrt(3)

#     # I_kG_contribution = I_kG.sum(axis=1)
#     # ratio_SG_ikss = I_kG_contribution / I_ikss
#     # close_to_SG = ratio_SG_ikss > 5e-2

#     close_to_SG = I_kG / I_rG > 2

#     if tk_s == 2e-2:
#         mu = 0.84 + 0.26 * np.exp(-0.26 * abs(I_kG) / I_rG)
#     elif tk_s == 5e-2:
#         mu = 0.71 + 0.51 * np.exp(-0.3 * abs(I_kG) / I_rG)
#     elif tk_s == 10e-2:
#         mu = 0.62 + 0.72 * np.exp(-0.32 * abs(I_kG) / I_rG)
#     elif tk_s >= 25e-2:
#         mu = 0.56 + 0.94 * np.exp(-0.38 * abs(I_kG) / I_rG)
#     else:
#         raise UserWarning('not implemented for other tk_s than 20ms, 50ms, 100ms and >=250ms')

#     mu = np.clip(mu, 0, 1)

#     I_ikss_G = abs(I_ikss - np.sum((1 - mu) * I_kG, axis=1))

#     # I_ikss_G = I_ikss - np.sum(abs(V_ikss.T[:, gen_buses_ppc]) * (1-mu) * I_kG, axis=1)

#     I_ikss_G = abs(I_ikss - np.sum(dV_G / V_Is * (1 - mu) * I_kG, axis=1))

#     return I_ikss_G


def _calc_branch_currents(net, ppci, bus_idx):
    n_sc_bus = np.shape(bus_idx)[0]

    case = net._options["case"]
    minmax = np.nanmin if case == "min" else np.nanmax

    Yf = ppci["internal"]["Yf"]
    Yt = ppci["internal"]["Yt"]
    baseI = ppci["internal"]["baseI"]
    n_bus = ppci["bus"].shape[0]
    fb = np.real(ppci["branch"][:, 0]).astype(np.int64)
    tb = np.real(ppci["branch"][:, 1]).astype(np.int64)

    # calculate voltage source branch current
    if net["_options"]["inverse_y"]:
        Zbus = ppci["internal"]["Zbus"]
        V_ikss = (ppci["bus"][:, IKSS1] * baseI) * Zbus
        V_ikss = V_ikss[:, bus_idx]
    else:
        ybus_fact = ppci["internal"]["ybus_fact"]
        V_ikss = np.zeros((n_bus, n_sc_bus), dtype=np.complex128)
        for ix, b in enumerate(bus_idx):
            ikss = np.zeros(n_bus, dtype=np.complex128)
            ikss[b] = ppci["bus"][b, IKSS1] * baseI[b]
            V_ikss[:, ix] = ybus_fact(ikss)

    ikss1_all_f = np.conj(Yf.dot(V_ikss))
    ikss1_all_t = np.conj(Yt.dot(V_ikss))
    ikss1_all_f[abs(ikss1_all_f) < 1e-10] = 0.
    ikss1_all_t[abs(ikss1_all_t) < 1e-10] = 0.

    # add current source branch current if there is one
    current_sources = any(~np.isnan(ppci["bus"][:, IKCV])) and np.any(ppci["bus"][:, IKCV] != 0)
    if current_sources:
        current = np.tile(-ppci["bus"][:, IKCV], (n_sc_bus, 1))
        for ix, b in enumerate(bus_idx):
            current[ix, b] += ppci["bus"][b, IKSS2]

        # calculate voltage source branch current
        if net["_options"]["inverse_y"]:
            Zbus = ppci["internal"]["Zbus"]
            V = np.dot((current * baseI), Zbus).T
        else:
            ybus_fact = ppci["internal"]["ybus_fact"]
            V = np.zeros((n_bus, n_sc_bus), dtype=np.complex128)
            for ix, b in enumerate(bus_idx):
                V[:, ix] = ybus_fact(current[ix, :] * baseI[b])

        fb = np.real(ppci["branch"][:, 0]).astype(np.int64)
        tb = np.real(ppci["branch"][:, 1]).astype(np.int64)
        ikss2_all_f = np.conj(Yf.dot(V))
        ikss2_all_t = np.conj(Yt.dot(V))

        ikss_all_f = abs(ikss1_all_f + ikss2_all_f)
        ikss_all_t = abs(ikss1_all_t + ikss2_all_t)
    else:
        ikss_all_f = abs(ikss1_all_f)
        ikss_all_t = abs(ikss1_all_t)

    if net._options["return_all_currents"]:
        ppci["internal"]["branch_ikss_f"] = ikss_all_f / baseI[fb, None]
        ppci["internal"]["branch_ikss_t"] = ikss_all_t / baseI[tb, None]
    else:
        ikss_all_f[abs(ikss_all_f) < 1e-10] = np.nan
        ikss_all_t[abs(ikss_all_t) < 1e-10] = np.nan
        ppci["branch"][:, IKSS_F] = np.nan_to_num(minmax(ikss_all_f, axis=1) / baseI[fb])
        ppci["branch"][:, IKSS_T] = np.nan_to_num(minmax(ikss_all_t, axis=1) / baseI[tb])

    if net._options["ip"]:
        kappa = ppci["bus"][:, KAPPA]
        if current_sources:
            ip_all_f = np.sqrt(2) * (ikss1_all_f * kappa[bus_idx] + ikss2_all_f)
            ip_all_t = np.sqrt(2) * (ikss1_all_t * kappa[bus_idx] + ikss2_all_t)
        else:
            ip_all_f = np.sqrt(2) * ikss1_all_f * kappa[bus_idx]
            ip_all_t = np.sqrt(2) * ikss1_all_t * kappa[bus_idx]

        if net._options["return_all_currents"]:
            ppci["internal"]["branch_ip_f"] = abs(ip_all_f) / baseI[fb, None]
            ppci["internal"]["branch_ip_t"] = abs(ip_all_t) / baseI[tb, None]
        else:
            ip_all_f[abs(ip_all_f) < 1e-10] = np.nan
            ip_all_t[abs(ip_all_t) < 1e-10] = np.nan
            ppci["branch"][:, IP_F] = np.nan_to_num(minmax(abs(ip_all_f), axis=1) / baseI[fb])
            ppci["branch"][:, IP_T] = np.nan_to_num(minmax(abs(ip_all_t), axis=1) / baseI[tb])

    if net._options["ith"]:
        n = 1
        m = ppci["bus"][bus_idx, M]
        ith_all_f = ikss_all_f * np.sqrt(m + n)
        ith_all_t = ikss_all_t * np.sqrt(m + n)

        if net._options["return_all_currents"]:
            ppci["internal"]["branch_ith_f"] = ith_all_f / baseI[fb, None]
            ppci["internal"]["branch_ith_t"] = ith_all_t / baseI[tb, None]
        else:
            ppci["branch"][:, ITH_F] = np.nan_to_num(minmax(ith_all_f, axis=1) / baseI[fb])
            ppci["branch"][:, ITH_T] = np.nan_to_num(minmax(ith_all_t, axis=1) / baseI[fb])

    # Update bus index for branch results
    if net._options["return_all_currents"]:
        ppci["internal"]["br_res_ks_ppci_bus"] = bus_idx


def nan_minmax(a, rows, argminmax):
    # because numpy won't sort complex values by magnitude :(
    rows_allnan = np.isnan(a[rows, :]).all(axis=1)
    aa = a[rows, :]
    minmax_a = np.zeros(a.shape[0], dtype=np.complex128)
    minmax_a[~rows_allnan] = aa[~rows_allnan, argminmax(np.abs(aa[~rows_allnan, :]), axis=1)]
    return minmax_a


def _calc_branch_currents_complex(net, ppci, bus_idx):
    net["ppci"] = ppci  #todo remove this
    n_sc_bus = np.shape(bus_idx)[0]

    case = net._options["case"]
    minmax = np.nanmin if case == "min" else np.nanmax
    argminmax = np.nanargmin if case == "min" else np.nanargmax

    ikss1 = ppci["bus"][:, IKSS1] * np.exp(1j * np.deg2rad(ppci["bus"][:, PHI_IKSS1_DEGREE]))
    V_ikss = ppci["internal"]["V_ikss"]
    valid_V = ppci["internal"]["valid_V"]

    Yf = ppci["internal"]["Yf"]
    Yt = ppci["internal"]["Yt"]
    baseI = ppci["internal"]["baseI"]
    n_bus = ppci["bus"].shape[0]
    fb = np.real(ppci["branch"][:, 0]).astype(np.int64)
    tb = np.real(ppci["branch"][:, 1]).astype(np.int64)

    # branch = ppci["branch"]
    # offtap = np.flatnonzero(branch[:, TAP] != 1)
    # if len(offtap) != 0:
    #     branch[offtap, BR_R] *= branch[offtap, TAP]
    #     branch[offtap, BR_X] *= branch[offtap, TAP]
    #     Ybus_corr, _, _ = makeYbus(ppci["baseMVA"], ppci["bus"], branch)
    #     Zbus = inv(Ybus_corr.toarray())
    # else:
    #     Zbus = ppci["internal"]["Zbus"]

    # # calculate voltage source branch current
    # if net["_options"]["inverse_y"]:
    #     Zbus = ppci["internal"]["Zbus"]
    #     ikss1 = ppci["bus"][:, IKSS1] * np.exp(1j * np.deg2rad(ppci["bus"][:, PHI_IKSS1_DEGREE]))
    #     # V_ikss1 = (ikss1 * baseI * Zbus)  # making it a complex calculation
    #     # V_ikss1 = V_ikss1[:, bus_idx]
    #     # V_ikss11 = V_ikss1[np.argmax(np.abs(V_ikss1), axis=0), np.arange(V_ikss1.shape[1])] - V_ikss1  # numpy indexing issue
    #     # todo explain formula
    #     V_ikss = V0 - Zbus[:, bus_idx] / Zbus[bus_idx, bus_idx] * V0[bus_idx]
    #     V_ikss[np.abs(V_ikss) < 1e-10] = 0
    #     # ikss1 = ppci["internal"]["Ybus"] * V_ikss
    #     V_ikss[0] / Zbus[bus_idx,bus_idx] / ppci["bus"][bus_idx, BASE_KV] / np.sqrt(3) * ppci["baseMVA"]
    #     print(abs(V_ikss), np.angle(V_ikss, deg=True))
    #     I0 = ppci["internal"]["Ybus"] * V0
    #     ikss1[bus_idx] = I0[bus_idx] - ikss1[bus_idx]
    #     abs(V0[0]) / Zbus[bus_idx, bus_idx] / ppci["bus"][bus_idx, BASE_KV] / np.sqrt(3) * ppci["baseMVA"]
    #
    # else:
    #     # todo
    #     ybus_fact = ppci["internal"]["ybus_fact"]
    #     V_ikss = np.zeros((n_bus, n_sc_bus), dtype=np.complex128)
    #     for ix, b in enumerate(bus_idx):
    #         ikss = np.zeros(n_bus, dtype=np.complex128)
    #         ikss[b] = ppci["bus"][b, IKSS1] * np.exp(1j * np.deg2rad(ppci["bus"][b, PHI_IKSS1_DEGREE])) * baseI[b]
    #         V_ikss[:, ix] = ybus_fact(ikss)

    # net_copy = net.deepcopy()
    # pp.runpp(net_copy)
    # Ybus_p = net_copy._ppc["internal"]["Ybus"]
    # Yf_p = net_copy._ppc["internal"]["Yf"]
    # Yt_ = net_copy._ppc["internal"]["Yt"]
    # from scipy.linalg import inv
    # Zbus_p = inv(Ybus_p.toarray())

    # V_ikss_p = (ikss1 * baseI * Zbus_p)[:, bus_idx]
    #

    # z = 1.05 / (ikss1 * baseI)[bus_idx]
    # # ppci["bus"][0, [4, 5]] = 0
    # ppci["bus"][bus_idx, 4] = z.real
    # ppci["bus"][bus_idx, 5] = z.imag
    # Ybus, Yf, Yt = makeYbus(ppci["baseMVA"], ppci["bus"], ppci["branch"])
    # Zbus = inv(Ybus.toarray())
    # V_ikss = (ikss1 * baseI * Zbus)[:, bus_idx]

    # if len(tap_branch) != 0:
    #     factor[tap_branch[:, F_BUS].real.astype(np.int64)] = tap_branch[:, TAP].real

    # branch = ppci["branch"]
    # offtap = np.flatnonzero(branch[:, TAP] != 1)
    # if len(offtap) != 0:
    #     branch[offtap, BR_R] *= branch[offtap, TAP]
    #     branch[offtap, BR_X] *= branch[offtap, TAP]
    #     Ybus_corr, Yf, Yt = makeYbus(ppci["baseMVA"], ppci["bus"], branch)
    #     Zbus = inv(Ybus_corr.toarray())
    # else:
    #     Zbus = ppci["internal"]["Zbus"]

    # V_ikss = V_ikss[np.argmax(np.abs(V_ikss), axis=0), np.arange(V_ikss.shape[1])] - V_ikss  # numpy indexing issue

    # V_ikss = V0 - V_ikss  # numpy indexing issue
    # V_ikss[bus_idx] = - 1.05  # numpy indexing issue
    # Ybus = ppci["internal"]["Ybus"]
    # diff_i = Ybus @ V_ikss
    # V2=((diff_i + ikss1 * baseI) * Zbus)[:, bus_idx]

    ikss1_all_f = Yf.dot(V_ikss)
    ikss1_all_t = Yt.dot(V_ikss)
    ikss1_all_f[np.abs(ikss1_all_f) < 1e-10] = 0.
    ikss1_all_t[np.abs(ikss1_all_t) < 1e-10] = 0.

    # add current source branch current if there is one
    current_sources = any(~np.isnan(ppci["bus"][:, IKCV])) and np.any(ppci["bus"][:, IKCV] != 0)
    if current_sources:
        ikcv = ppci["bus"][:, IKCV] * np.exp(np.deg2rad(ppci["bus"][:, PHI_IKCV_DEGREE]) * 1j)
        ikss2 = ppci["bus"][:, IKSS2] * np.exp(1j * np.deg2rad(ppci["bus"][:, PHI_IKSS2_DEGREE]))
        current = np.tile(ikcv, (n_sc_bus, 1))
        # np.fill_diagonal(current, current.diagonal() - ikss2)
        for ix, b in enumerate(bus_idx):
            # current[ix, b] -= ppci["bus"][b, IKSS2] * np.exp(np.deg2rad(ppci["bus"][b, PHI_IKSS2_DEGREE])*1j)
            current[ix, b] -= ikss2[b]

        # calculate voltage source branch current
        if net["_options"]["inverse_y"]:
            Zbus = ppci["internal"]["Zbus"]
            V = np.dot(Zbus, (current * baseI).T)
        else:
            ybus_fact = ppci["internal"]["ybus_fact"]
            V = np.zeros((n_bus, n_sc_bus), dtype=np.complex128)
            for ix, b in enumerate(bus_idx):
                V[:, ix] = ybus_fact(current[ix, :] * baseI)

        fb = np.real(ppci["branch"][:, 0]).astype(np.int64)
        tb = np.real(ppci["branch"][:, 1]).astype(np.int64)

        V[np.abs(V) < 1e-10] = 0

        ikss2_all_f = Yf.dot(V)
        ikss2_all_t = Yt.dot(V)
        ikss2_all_f[np.abs(ikss2_all_f) < 1e-10] = 0.
        ikss2_all_t[np.abs(ikss2_all_t) < 1e-10] = 0.

        V_ikss += V  # superposition

        # ikss_all_f = Yf.dot(V_ikss)
        # ikss_all_t = Yt.dot(V_ikss)
        ikss_all_f = ikss1_all_f + ikss2_all_f
        ikss_all_t = ikss1_all_t + ikss2_all_t

    else:

        ikss_all_f = ikss1_all_f
        ikss_all_t = ikss1_all_t

    # ikss_all_f[np.abs(ikss_all_f) < 1e-10] = np.nan
    # ikss_all_t[np.abs(ikss_all_t) < 1e-10] = np.nan

    # calculate active and reactive power, voltages
    skss_all_f = np.conj(ikss_all_f) * V_ikss[fb]
    pkss_all_f = skss_all_f.real
    qkss_all_f = skss_all_f.imag

    skss_all_t = np.conj(ikss_all_t) * V_ikss[tb]
    pkss_all_t = skss_all_t.real
    qkss_all_t = skss_all_t.imag

    rows_fb = np.arange(len(fb))
    rows_tb = np.arange(len(tb))
    minmax_vkss_all_f = V_ikss[fb][rows_fb, argminmax(np.abs(V_ikss[fb]), axis=1)].flatten()
    minmax_vkss_all_t = V_ikss[tb][rows_tb, argminmax(np.abs(V_ikss[tb]), axis=1)].flatten()

    ikss_all_f[abs(ikss_all_f) < 1e-10] = np.nan
    ikss_all_t[abs(ikss_all_t) < 1e-10] = np.nan
    # minmax_ikss_all_f = ikss_all_f[rows_fb, argminmax(np.abs(ikss_all_f), axis=1)].flatten()
    # minmax_ikss_all_t = ikss_all_t[rows_tb, argminmax(np.abs(ikss_all_t), axis=1)].flatten()
    minmax_ikss_all_f = nan_minmax(ikss_all_f, rows_fb, argminmax)
    minmax_ikss_all_t = nan_minmax(ikss_all_t, rows_tb, argminmax)

    if net._options["return_all_currents"]:
        ppci["internal"]["branch_ikss_f"] = np.nan_to_num(np.abs(ikss_all_f)) / baseI[fb, None]
        ppci["internal"]["branch_ikss_t"] = np.nan_to_num(np.abs(ikss_all_t)) / baseI[tb, None]

        ppci["internal"]["branch_ikss_angle_f"] = np.nan_to_num(np.angle(ikss_all_f, deg=True))
        ppci["internal"]["branch_ikss_angle_t"] = np.nan_to_num(np.angle(ikss_all_t, deg=True))

        if valid_V:
            ppci["internal"]["branch_pkss_f"] = np.nan_to_num(pkss_all_f) * ppci["baseMVA"]
            ppci["internal"]["branch_pkss_t"] = np.nan_to_num(pkss_all_t) * ppci["baseMVA"]

            ppci["internal"]["branch_qkss_f"] = np.nan_to_num(qkss_all_f) * ppci["baseMVA"]
            ppci["internal"]["branch_qkss_t"] = np.nan_to_num(qkss_all_t) * ppci["baseMVA"]

            ppci["internal"]["branch_vkss_f"] = np.nan_to_num(np.abs(V_ikss[fb]))
            ppci["internal"]["branch_vkss_t"] = np.nan_to_num(np.abs(V_ikss[tb]))

            ppci["internal"]["branch_vkss_angle_f"] = np.nan_to_num(np.angle(V_ikss[fb], deg=True))
            ppci["internal"]["branch_vkss_angle_t"] = np.nan_to_num(np.angle(V_ikss[tb], deg=True))
    else:
        # ikss_all_f[np.abs(ikss_all_f) < 1e-10] = np.nan
        # ikss_all_t[np.abs(ikss_all_t) < 1e-10] = np.nan
        ppci["branch"][:, IKSS_F] = np.abs(minmax_ikss_all_f) / baseI[fb]
        ppci["branch"][:, IKSS_ANGLE_F] = np.angle(minmax_ikss_all_f, deg=True)
        ppci["branch"][:, IKSS_T] = np.abs(minmax_ikss_all_t) / baseI[tb]
        ppci["branch"][:, IKSS_ANGLE_T] = np.angle(minmax_ikss_all_t, deg=True)

        if valid_V:
            ppci["branch"][:, PKSS_F] = np.nan_to_num(minmax(pkss_all_f, axis=1)) * ppci["baseMVA"]
            ppci["branch"][:, QKSS_F] = np.nan_to_num(minmax(qkss_all_f, axis=1)) * ppci["baseMVA"]

            ppci["branch"][:, PKSS_T] = np.nan_to_num(minmax(pkss_all_t, axis=1)) * ppci["baseMVA"]
            ppci["branch"][:, QKSS_T] = np.nan_to_num(minmax(qkss_all_t, axis=1)) * ppci["baseMVA"]

            ppci["branch"][:, VKSS_MAGN_F] = np.abs(minmax_vkss_all_f)
            ppci["branch"][:, VKSS_MAGN_T] = np.abs(minmax_vkss_all_t)

            ppci["branch"][:, VKSS_ANGLE_F] = np.angle(minmax_vkss_all_f, deg=True)
            ppci["branch"][:, VKSS_ANGLE_T] = np.angle(minmax_vkss_all_t, deg=True)

    if net._options["ip"]:
        kappa = ppci["bus"][:, KAPPA]
        if current_sources:
            ip_all_f = np.sqrt(2) * (ikss1_all_f * kappa[bus_idx] + ikss2_all_f)
            ip_all_t = np.sqrt(2) * (ikss1_all_t * kappa[bus_idx] + ikss2_all_t)
        else:
            ip_all_f = np.sqrt(2) * ikss1_all_f * kappa[bus_idx]
            ip_all_t = np.sqrt(2) * ikss1_all_t * kappa[bus_idx]

        if net._options["return_all_currents"]:
            ppci["internal"]["branch_ip_f"] = np.nan_to_num(np.abs(ip_all_f)) / baseI[fb, None]
            ppci["internal"]["branch_ip_t"] = np.nan_to_num(np.abs(ip_all_t)) / baseI[tb, None]
        else:
            ip_all_f[np.abs(ip_all_f) < 1e-10] = np.nan
            ip_all_t[np.abs(ip_all_t) < 1e-10] = np.nan
            # ppci["branch"][:, IP_F] = minmax(np.abs(ip_all_f), axis=1) / baseI[fb]
            # ppci["branch"][:, IP_T] = minmax(np.abs(ip_all_t), axis=1) / baseI[tb]
            ppci["branch"][:, IP_F] = np.nanmax(np.abs(ip_all_f), axis=1) / baseI[fb]
            ppci["branch"][:, IP_T] = np.nanmax(np.abs(ip_all_t), axis=1) / baseI[tb]

    if net._options["ith"]:
        n = 1
        m = ppci["bus"][bus_idx, M]
        ith_all_f = np.abs(ikss_all_f * np.sqrt(m + n))
        ith_all_t = np.abs(ikss_all_t * np.sqrt(m + n))

        if net._options["return_all_currents"]:
            ppci["internal"]["branch_ith_f"] = np.nan_to_num(np.abs(ith_all_f)) / baseI[fb, None]
            ppci["internal"]["branch_ith_t"] = np.nan_to_num(np.abs(ith_all_t)) / baseI[tb, None]
        else:
            ith_all_f[np.abs(ith_all_f) < 1e-10] = np.nan
            ith_all_t[np.abs(ith_all_t) < 1e-10] = np.nan
            # ppci["branch"][:, ITH_F] = minmax(np.abs(ith_all_f), axis=1) / baseI[fb]
            # ppci["branch"][:, ITH_T] = minmax(np.abs(ith_all_t), axis=1) / baseI[fb]
            ppci["branch"][:, ITH_F] = np.nanmax(np.abs(ith_all_f), axis=1) / baseI[fb]
            ppci["branch"][:, ITH_T] = np.nanmax(np.abs(ith_all_t), axis=1) / baseI[fb]

    # Update bus index for branch results
    if net._options["return_all_currents"]:
        ppci["internal"]["br_res_ks_ppci_bus"] = bus_idx



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np
from scipy.stats import chi2

from pandapower.auxiliary import version_check
try:
    from numba import jit
    version_check('numba')
except ImportError:
    from pandapower.pf.no_numba import jit


from pandapower.estimation.algorithm.matrix_base import BaseAlgebra
from pandapower.estimation.ppc_conversion import ExtendedPPCI


def get_estimator(base_class, estimator_name):
    assert base_class in (BaseEstimatorIRWLS, BaseEstimatorOpt)

    available_estimators = {estm_cls.__name__.split("Estimator")[0].lower(): estm_cls
                            for estm_cls in base_class.__subclasses__()}
    if estimator_name.lower() not in available_estimators:
        raise Exception("Estimator not available! Try another one!")
    else:
        return available_estimators[estimator_name.lower()]


class BaseEstimatorIRWLS(BaseAlgebra):
    def __init__(self, eppci: ExtendedPPCI, **hyperparameters):
        # Initilize BaseAlgebra object for calculation of relevant matrix
        super(BaseEstimatorIRWLS, self).__init__(eppci)

    def create_phi(self, E):  # pragma: no cover
        # Must be implemented for subclasses!
        pass


class BaseEstimatorOpt(BaseAlgebra):
    def __init__(self, eppci, **hyperparameters):
        super(BaseEstimatorOpt, self).__init__(eppci)
        # Hyperparameters for estimator should be added here

    def cost_function(self, E):  # pragma: no cover
        # Minimize sum(cost(r))
        # r = cost(z - h(x))
        # Must be implemented according to the estimator for the optimization
        pass

    def create_cost_jacobian(self, E):  # pragma: no cover
        pass


class WLSEstimator(BaseEstimatorOpt, BaseEstimatorIRWLS):
    def __init__(self, eppci, **hyperparameters):
        super(WLSEstimator, self).__init__(eppci, **hyperparameters)

    def cost_function(self, E):
        rx = self.create_rx(E)
        cost = np.sum((1/self.sigma**2) * (rx**2))
        return cost

    def create_cost_jacobian(self, E):
        # dr/dE = drho / dr * d(z-hx) / dE
        # dr/dE = (drho/dr) * - (d(hx)/dE)
        # 2 * rx * (1/sigma**2)* -(dhx/dE)
        rx = self.create_rx(E)
        hx_jac = self.create_hx_jacobian(E)
        drho_dr = 2 * (rx * (1/self.sigma**2))
        jac = - np.sum(drho_dr.reshape((-1, 1)) * hx_jac, axis=0)
        return jac

    def create_phi(self, E):
        # Standard WLS does not update this matrix
        return np.diagflat(1/self.sigma**2)


class SHGMEstimatorIRWLS(BaseEstimatorIRWLS):
    def __init__(self, eppci: ExtendedPPCI, **hyperparameters):
        super(SHGMEstimatorIRWLS, self).__init__(eppci, **hyperparameters)
        assert 'a' in hyperparameters
        self.a = hyperparameters.get('a')

    def create_phi(self, E):
        r = self.create_rx(E)
        _, w = self.weight(E)
        rsi = r / (w * self.sigma)
        phi = 1/(self.sigma**2)
        condition_mask = np.abs(rsi)>self.a
        phi[condition_mask] = (1/(self.sigma**2) * np.abs(self.a / rsi))[condition_mask]
        return np.diagflat(phi)

    def weight(self, E):
        H = self.create_hx_jacobian(E)
        v = np.sum(H != 0, axis=1)
        chi2_res = chi2.ppf(0.975, v)
        ps = self._ps(H)
        return chi2_res, np.min(np.c_[(chi2_res/ps)**2, np.ones(ps.shape)], axis=1)

    @staticmethod
    def _ps(H):
        omega = np.dot(H, H.T)

        x = np.zeros(omega.shape[0]-1)
        y = np.zeros(omega.shape[0])
        sm = np.zeros(omega.shape[0])
        ps = np.zeros(omega.shape[0])

        @jit
        def calc_sm(omega, x, y, sm):  # pragma: no cover
            m = omega.shape[0]
            x_shape = x.shape[0]
            y_shape = y.shape[0]
            count0 = 0
            for k in range(m):
                for i in range(m):
                    count0 = 0
                    for j in range(m):
                        if j != i:
                            x_ix = j if j < i else j-1
                            x[x_ix] = np.abs(omega[i, k]+omega[j, k])
                            if not x[x_ix]:
                                count0 += 1
                    x.sort()
                    y[i] = x[count0 + (x_shape - count0 + 1)//2 - 1]
                y.sort()
                sm[k] = y[(y_shape + 1)//2 - 1] * 1.1926
            return sm

        sm = calc_sm(omega, x, y, sm)
        for i in range(omega.shape[0]):
            ps[i] = np.max(np.abs(omega[i, :])/sm)
        return ps


class LAVEstimator(BaseEstimatorOpt):
    def cost_function(self, E):
        rx = self.create_rx(E)
        cost = np.sum(np.abs(rx))
        return cost

    def create_cost_jacobian(self, E):
        # dr/dE = drho / dr * d(z-hx) / dE
        # dr/dE = (drho/dr) * - (d(hx)/dE)
        # sign(rx) * -(dhx/dE)
        rx = self.create_rx(E)
        hx_jac = self.create_hx_jacobian(E)
        drho_dr = np.sign(rx)
        jac = - np.sum(drho_dr.reshape((-1, 1)) * hx_jac, axis=0)
        return jac


class QCEstimatorOpt(BaseEstimatorOpt):
    def __init__(self, eppci, **hyperparameters):
        super(QCEstimatorOpt, self).__init__(eppci, **hyperparameters)
        assert 'a' in hyperparameters
        self.a = hyperparameters['a']

    def cost_function(self, E):
        rx = self.create_rx(E)
        cost = (1/self.sigma**2) * (rx**2)
        large_dev_mask = np.abs(rx/self.sigma) > self.a
        if np.any(large_dev_mask):
            cost[large_dev_mask] = (self.a**2)
        return np.sum(cost)

    def create_cost_jacobian(self, E):
        # dr/dE = drho / dr * d(z-hx) / dE
        # dr/dE = (drho/dr) * - (d(hx)/dE)
        # 2 * rx * -(dhx/dE) if np.abs(rx/sigma) < a
        # 0 else
        rx = self.create_rx(E)
        hx_jac = self.create_hx_jacobian(E)
        drho_dr = 2 * (rx * (1/self.sigma)**2)
        large_dev_mask = (np.abs(rx/self.sigma) > self.a)
        if np.any(large_dev_mask):
            drho_dr[large_dev_mask] = 0.001
        jac = - np.sum(drho_dr.reshape((-1, 1)) * hx_jac, axis=0)
        return jac


class QLEstimatorOpt(BaseEstimatorOpt):
    def __init__(self, eppci, **hyperparameters):
        super(QLEstimatorOpt, self).__init__(eppci, **hyperparameters)
        assert 'a' in hyperparameters
        self.a = hyperparameters['a']

    def cost_function(self, E):
        rx = self.create_rx(E)
        cost = (1/self.sigma**2) * (rx**2)
        large_dev_mask = (np.abs(rx/self.sigma) > self.a)
        if np.any(large_dev_mask):
            cost[large_dev_mask] = (np.abs(rx/self.sigma) - self.a +
                self.a**2)[large_dev_mask]
        return np.sum(cost)

    def create_cost_jacobian(self, E):
        # dr/dE = drho / dr * d(z-hx) / dE
        # dr/dE = (drho/dr) * - (d(hx)/dE)
        # 2 * rx * -(dhx/dE) if np.abs(rx/sigma) < a
        # 2 * drho/dr * -(dhx/dE) else
        rx = self.create_rx(E)
        hx_jac = self.create_hx_jacobian(E)
        drho_dr = 2 * (rx * (1/self.sigma)**2)
        large_dev_mask = np.abs(rx/self.sigma) > self.a
        if np.any(large_dev_mask):
            drho_dr[large_dev_mask] = (np.sign(rx)* (1/self.sigma))[large_dev_mask]
        jac = - np.sum(drho_dr.reshape((-1, 1)) * hx_jac, axis=0)
        return jac


import logging
import time

import pandas as pd
import numpy as np

from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net
from pandapower.converter.cim.other_classes import Report, LogLevel, ReportCode

logger = logging.getLogger('cim.cim2pp.converter_classes.linearShuntCompensatorCim16')

sc = cim_tools.get_pp_net_special_columns_dict()


class LinearShuntCompensatorCim16:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def convert_linear_shunt_compensator_cim16(self):
        time_start = time.time()
        self.logger.info("Start converting LinearShuntCompensator.")
        eqssh_shunts = self._prepare_linear_shunt_compensator_cim16()
        self.cimConverter.copy_to_pp('shunt', eqssh_shunts)
        self.logger.info("Created %s shunts in %ss." % (eqssh_shunts.index.size, time.time() - time_start))
        self.cimConverter.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
            message="Created %s shunts from LinearShuntCompensator in %ss." %
                    (eqssh_shunts.index.size, time.time() - time_start)))

    def _prepare_linear_shunt_compensator_cim16(self) -> pd.DataFrame:
        eqssh_shunts = self.cimConverter.merge_eq_ssh_profile('LinearShuntCompensator', add_cim_type_column=True)
        eqssh_shunts = pd.merge(eqssh_shunts, self.cimConverter.bus_merge, how='left', on='rdfId')
        if 'inService' in eqssh_shunts.columns:
            eqssh_shunts['connected'] = eqssh_shunts['connected'] & eqssh_shunts['inService']
        eqssh_shunts = eqssh_shunts.rename(columns={
            'rdfId': sc['o_id'], 'rdfId_Terminal': sc['t'], 'connected': 'in_service', 'index_bus': 'bus',
            'nomU': 'vn_kv', 'sections': 'step', 'maximumSections': 'max_step'})
        # complex conductance
        y = eqssh_shunts['gPerSection'] + eqssh_shunts['bPerSection'] * 1j
        # apparent power
        s = eqssh_shunts['vn_kv'] ** 2 * np.conj(y)
        eqssh_shunts['p_mw'] = s.values.real
        eqssh_shunts['q_mvar'] = s.values.imag
        # create step_dependency_table flag
        if 'step_dependency_table' not in eqssh_shunts.columns:
            eqssh_shunts["step_dependency_table"] = False
        return eqssh_shunts



import os
from pandapower.converter.pandamodels.to_pm import convert_to_pm_structure, dump_pm_json
from pandapower.converter.pandamodels.from_pm import read_pm_results_to_net
from pandapower.optimal_powerflow import OPFNotConverged

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging
logger = logging.getLogger(__name__)


def _runpm(net, delete_buffer_file=True, pm_file_path=None, pdm_dev_mode=False, **kwargs): 
    """
    Converts the pandapower net to a pm json file, saves it to disk, runs a PandaModels.jl, and reads
    the results back to the pandapower net:
    INPUT
    ----------
    **net** - pandapower net
    OPTIONAL
    ----------
    **delete_buffer_file** (bool, True) - deletes the pm buffer json file if True.
    **pm_file_path** -path to save the converted net json file.
    **pdm_dev_mode** (bool, False) - If True, the develop mode of PdM is called.
    """
    # convert pandapower to power models file -> this is done in python
    net, pm, ppc, ppci = convert_to_pm_structure(net, **kwargs)
    # call optional callback function
    if net._options["pp_to_pm_callback"] is not None:
        net._options["pp_to_pm_callback"](net, ppci, pm)
    # writes pm json to disk, which is loaded afterwards in julia
    buffer_file = dump_pm_json(pm, pm_file_path)
    logger.debug("the json file for converted net is stored in: %s" % buffer_file)
    # run power models optimization in julia
    result_pm = _call_pandamodels(buffer_file, net._options["julia_file"], pdm_dev_mode)
    
    logger.info("Optimization ('"+net._options["julia_file"]+"') " +
                "is finished in %s seconds:" % round(result_pm["solve_time"], 2))
    # read results and write back to net
    try:
        read_pm_results_to_net(net, ppc, ppci, result_pm)
        if pm_file_path is None and delete_buffer_file:
            # delete buffer file after calculation
            os.remove(buffer_file)
            logger.debug("the json file for converted net is deleted from %s" % buffer_file)        
    except OPFNotConverged as e:
        if pm_file_path is None and delete_buffer_file:
            os.remove(buffer_file)
            logger.debug("the json file for converted net is deleted from %s" % buffer_file)
        raise e
    except Exception as e:
        raise e

def _call_pandamodels(buffer_file, julia_file, dev_mode):  # pragma: no cover

    try:
        import julia
        from julia import Main
        from julia import Pkg
        from julia import Base
    except ImportError:
        raise ImportError(
            "Please install pyjulia properly to run pandapower with PandaModels.jl.")
        
    try:
        julia.Julia()
    except:
        raise UserWarning(
            "Could not connect to julia, please check that Julia is installed and pyjulia is correctly configured")
              
    if not Base.find_package("PandaModels"):
        logger.info("PandaModels.jl is not installed in julia. It is added now!")
        Pkg.Registry.update()
        Pkg.add("PandaModels")  
        
        if dev_mode:
            logger.info("installing dev mode is a slow process!")
            Pkg.resolve()
            Pkg.develop("PandaModels")
            # add pandamodels dependencies: slow process
            Pkg.instantiate()
            
        Pkg.build()
        Pkg.resolve()
        logger.info("Successfully added PandaModels")

    if dev_mode:
        Pkg.develop("PandaModels")
        Pkg.build()
        Pkg.resolve()
        Pkg.activate("PandaModels")

    try:
        Main.using("PandaModels")
    except ImportError:
        raise ImportError("cannot use PandaModels")

    Main.buffer_file = buffer_file
    result_pm = Main.eval(julia_file + "(buffer_file)")

    # if dev_mode:
    #     Pkg.activate()
    #     Pkg.free("PandaModels")
    #     Pkg.resolve()
    return result_pm


# # just for you delete it later
# def _call_powermodels(buffer_file, julia_file):  # pragma: no cover
#     # checks if julia works, otherwise raises an error
#     try:
#         import julia
#         from julia import Main
#     except ImportError:
#         raise ImportError("Please install pyjulia to run pandapower with PowerModels.jl")
#     try:
#         j = julia.Julia()
#     except:
#         raise UserWarning(
#             "Could not connect to julia, please check that Julia is installed and pyjulia is correctly configured")

#     # import two julia scripts and runs powermodels julia_file
#     # Main.include(os.path.join(pp_dir, "opf", 'pp_2_pm.jl'))
#     try:
#         run_powermodels = Main.include(julia_file)
#     except ImportError:
#         raise UserWarning("File %s could not be imported" % julia_file)
#     result_pm = run_powermodels(buffer_file)
#     return result_pm


from pandapower.networks.dickert_lv_networks import *
from pandapower.networks.synthetic_voltage_control_lv_networks import *
from pandapower.networks.cigre_networks import *
from pandapower.networks.create_examples import *
from pandapower.networks.kerber_extreme_networks import *
from pandapower.networks.kerber_networks import *
from pandapower.networks.mv_oberrhein import *
from pandapower.networks.power_system_test_cases import *
from pandapower.networks.simple_pandapower_test_networks import *
from pandapower.networks.ieee_europen_lv_asymmetric import *
from pandapower.networks.lv_schutterwald import *



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import logging
import time
from typing import Union, List, Type, Dict

from pandapower.auxiliary import pandapowerNet
from . import build_pp_net
from . import converter_classes as std_converter_classes
from .. import cim_classes
from .. import interfaces

logger = logging.getLogger('cim.cim2pp.from_cim')


def from_cim_dict(cim_parser: cim_classes.CimParser, log_debug=False, convert_line_to_switch: bool = False,
                  line_r_limit: float = 0.1, line_x_limit: float = 0.1,
                  repair_cim: Union[str, interfaces.CIMRepair] = None,
                  repair_cim_class: Type[interfaces.CIMRepair] = None,
                  repair_pp: Union[str, interfaces.PandapowerRepair] = None,
                  repair_pp_class: Type[interfaces.PandapowerRepair] = None,
                  custom_converter_classes: Dict = None,
                  **kwargs) -> pandapowerNet:
    """
    Creates a pandapower net from a CIM data structure.

    :param cim_parser: The CimParser with parsed cim data.
    :param log_debug: Set this parameter to True to enable logging at debug level. Optional, default: False
    :param convert_line_to_switch: Set this parameter to True to enable line -> switch conversion. All lines with a
        resistance lower or equal than line_r_limit or a reactance lower or equal than line_x_limit will become a
        switch. Optional, default: False
    :param line_r_limit: The limit from resistance. Optional, default: 0.1
    :param line_x_limit: The limit from reactance. Optional, default: 0.1
    :param repair_cim: The CIMRepair object or a path to its serialized object. Optional, default: None
    :param repair_cim_class: The CIMRepair class. Optional, default: None
    :param repair_pp: The PandapowerRepair object or a path to its serialized object. Optional, default: None
    :param repair_pp_class: The PandapowerRepair class. Optional, default: None
    :param custom_converter_classes: Dict to inject classes for different functionality. Optional, default: None
    :return: The pandapower net.
    """
    converter_classes = get_converter_classes()
    if custom_converter_classes is not None:
        for key in custom_converter_classes:
            converter_classes[key] = custom_converter_classes.get(key)

    # repair the input CIM data
    if repair_cim is not None and repair_cim_class is not None:
        repair_cim = repair_cim_class().deserialize(repair_cim, report_container=cim_parser.get_report_container())
        repair_cim.repair(cim_parser.get_cim_dict(), report_container=cim_parser.get_report_container())

    cim_converter = build_pp_net.CimConverter(cim_parser=cim_parser, converter_classes=converter_classes, **kwargs)
    pp_net = cim_converter.convert_to_pp(convert_line_to_switch=convert_line_to_switch, line_r_limit=line_r_limit,
                                         line_x_limit=line_x_limit, log_debug=log_debug, **kwargs)

    # repair the output pandapower network
    if repair_pp is not None and repair_pp_class is not None:
        repair_pp = repair_pp_class().deserialize(repair_pp, report_container=cim_parser.get_report_container())
        repair_pp.repair(pp_net, report_container=cim_parser.get_report_container())

    return pp_net


def get_converter_classes():
    converter_classes: Dict[str, classmethod] = {
        'ConnectivityNodesCim16': std_converter_classes.connectivitynodes.connectivityNodesCim16.ConnectivityNodesCim16,
        'externalNetworkInjectionsCim16':
            std_converter_classes.externalnetworks.externalNetworkInjectionsCim16.ExternalNetworkInjectionsCim16,
        'acLineSegmentsCim16': std_converter_classes.lines.acLineSegmentsCim16.AcLineSegmentsCim16,
        'dcLineSegmentsCim16': std_converter_classes.lines.dcLineSegmentsCim16.DcLineSegmentsCim16,
        'switchesCim16': std_converter_classes.switches.switchesCim16.SwitchesCim16,
        'energyConcumersCim16': std_converter_classes.loads.energyConcumersCim16.EnergyConsumersCim16,
        'conformLoadsCim16': std_converter_classes.loads.conformLoadsCim16.ConformLoadsCim16,
        'nonConformLoadsCim16': std_converter_classes.loads.nonConformLoadsCim16.NonConformLoadsCim16,
        'stationSuppliesCim16': std_converter_classes.loads.stationSuppliesCim16.StationSuppliesCim16,
        'synchronousMachinesCim16': std_converter_classes.generators.synchronousMachinesCim16.SynchronousMachinesCim16,
        'asynchronousMachinesCim16':
            std_converter_classes.generators.asynchronousMachinesCim16.AsynchronousMachinesCim16,
        'energySourcesCim16': std_converter_classes.generators.energySourcesCim16.EnergySourceCim16,
        'linearShuntCompensatorCim16':
            std_converter_classes.shunts.linearShuntCompensatorCim16.LinearShuntCompensatorCim16,
        'nonLinearShuntCompensatorCim16':
            std_converter_classes.shunts.nonLinearShuntCompensatorCim16.NonLinearShuntCompensatorCim16,
        'staticVarCompensatorCim16': std_converter_classes.shunts.staticVarCompensatorCim16.StaticVarCompensatorCim16,
        'equivalentBranchesCim16': std_converter_classes.impedance.equivalentBranchesCim16.EquivalentBranchesCim16,
        'seriesCompensatorsCim16': std_converter_classes.impedance.seriesCompensatorsCim16.SeriesCompensatorsCim16,
        'equivalentInjectionsCim16': std_converter_classes.wards.equivalentInjectionsCim16.EquivalentInjectionsCim16,
        'powerTransformersCim16': std_converter_classes.transformers.powerTransformersCim16.PowerTransformersCim16,
        'tapController': std_converter_classes.transformers.tapController.TapController,
        'geoCoordinatesFromGLCim16':
            std_converter_classes.coordinates.geoCoordinatesFromGLCim16.GeoCoordinatesFromGLCim16,
        'coordinatesFromDLCim16': std_converter_classes.coordinates.coordinatesFromDLCim16.CoordinatesFromDLCim16,
    }
    return converter_classes


def from_cim(file_list: List[str] = None, encoding: str = None, convert_line_to_switch: bool = False,
             line_r_limit: float = 0.1, line_x_limit: float = 0.1,
             repair_cim: Union[str, interfaces.CIMRepair] = None,
             repair_cim_class: Type[interfaces.CIMRepair] = None,
             repair_pp: Union[str, interfaces.PandapowerRepair] = None,
             repair_pp_class: Type[interfaces.PandapowerRepair] = None,
             custom_converter_classes: Dict = None,
             cgmes_version: str = '2.4.15', **kwargs) -> pandapowerNet:
    """
    Converts a CIM net to a pandapower net from XML files.
    Additional parameters for kwargs:
    - create_measurements (str): Set this parameter to 'SV' to create measurements for the pandapower net from the SV
    profile. Set it to 'Analog' to create measurements from Analogs. If the parameter is not set or is set to None, no
    measurements will be created.
    - diagram_name (str): The name from the Diagram from the DL profile for the diagram coordinates. Default:
    The first diagram sorted ascending by name. Set the parameter to "all" to use available diagrams for creating the
    coordinates.
    - create_tap_controller (bool): If True, create pandapower controllers for transformer tap changers. If False, skip
    creating them. Default: True
    - sn_mva (float): Set the sn_mva from the pandapower net to a specific value. This value is not given in CGMES.
    Default: None (pandapower default will be chosen)
    - run_powerflow (bool): Option to run to powerflow inside the converter to create res tables directly.
    Default: False.
    - ignore_errors (bool): Option to disable raising of internal errors. Useful if you need to get a network not matter
    if there are errors in the conversion. Default: True.

    :param file_list: The path to the CGMES files as a list.
    :param encoding: The encoding from the files. Optional, default: None
    :param convert_line_to_switch: Set this parameter to True to enable line -> switch conversion. All lines with a
        resistance lower or equal than line_r_limit or a reactance lower or equal than line_x_limit will become a
        switch. Optional, default: False
    :param line_r_limit: The limit from resistance. Optional, default: 0.1
    :param line_x_limit: The limit from reactance. Optional, default: 0.1
    :param repair_cim: The CIMRepair object or a path to its serialized object. Optional, default: None
    :param repair_cim_class: The CIMRepair class. Optional, default: None
    :param repair_pp: The PandapowerRepair object or a path to its serialized object. Optional, default: None
    :param repair_pp_class: The PandapowerRepair class. Optional, default: None
    :param custom_converter_classes: Dict to inject classes for different functionality. Optional, default: None
    :param cgmes_version: The CGMES version of the files, can be 3.0 or 2.4.15. Optional, default: 2.4.15
    :return: The pandapower net.
    """
    time_start_parsing = time.time()

    cim_parser = cim_classes.CimParser(cgmes_version=cgmes_version, **kwargs)
    cim_parser.parse_files(file_list=file_list, encoding=encoding, prepare_cim_net=True, set_data_types=True)

    time_start_converting = time.time()
    pp_net = from_cim_dict(cim_parser, convert_line_to_switch=convert_line_to_switch,
                           line_r_limit=line_r_limit, line_x_limit=line_x_limit, repair_cim=repair_cim,
                           repair_cim_class=repair_cim_class, repair_pp=repair_pp, repair_pp_class=repair_pp_class,
                           custom_converter_classes=custom_converter_classes, **kwargs)
    time_end_converting = time.time()
    logger.info("The pandapower net: \n%s" % pp_net)

    logger.info("Needed time for parsing: %s" % (time_start_converting - time_start_parsing))
    logger.info("Needed time for converting: %s" % (time_end_converting - time_start_converting))
    logger.info("Total Time: %s" % (time_end_converting - time_start_parsing))

    return pp_net
